#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from pyscf import lib
from pyscf.pbc import dft
from pyscf.pbc.tdscf import kuhf
from pyscf.pbc.tdscf.krks import _rebuild_df


class TDA(kuhf.TDA):
    def kernel(self, x0=None):
        _rebuild_df(self)
        kuhf.TDA.kernel(self, x0=x0)

KTDA = TDA

class TDDFT(kuhf.TDHF):
    def kernel(self, x0=None):
        _rebuild_df(self)
        kuhf.TDHF.kernel(self, x0=x0)

RPA = KTDDFT = TDDFT

dft.kuks.KUKS.TDA   = lib.class_as_method(KTDA)
dft.kuks.KUKS.TDHF  = None
dft.kuks.KUKS.TDDFT = lib.class_as_method(TDDFT)



# Copyright 2023 The GPU4PySCF Authors. All Rights Reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Gradient of SMD solvent model, copied from GPU4PySCF with modification for CPU
'''
# pylint: disable=C0103

import numpy as np
from pyscf import lib
from pyscf.grad import rhf as rhf_grad

from pyscf.solvent import smd
from pyscf.solvent.grad import pcm as pcm_grad
from pyscf.lib import logger

def grad_solver(pcmobj, dm):
    '''
    dE = 0.5*v* d(K^-1 R) *v + q*dv
    v^T* d(K^-1 R)v = v^T*K^-1(dR - dK K^-1R)v = v^T K^-1(dR - dK q)
    '''
    mol = pcmobj.mol
    log = logger.new_logger(mol, mol.verbose)
    t1 = (logger.process_clock(), logger.perf_counter())
    if not pcmobj._intermediates:
        pcmobj.build()
    dm_cache = pcmobj._intermediates.get('dm', None)
    if dm_cache is not None and np.linalg.norm(dm_cache - dm) < 1e-10:
        pass
    else:
        pcmobj._get_vind(dm)

    gridslice    = pcmobj.surface['gslice_by_atom']
    v_grids      = pcmobj._intermediates['v_grids']
    A            = pcmobj._intermediates['A']
    D            = pcmobj._intermediates['D']
    S            = pcmobj._intermediates['S']
    K            = pcmobj._intermediates['K']
    q            = pcmobj._intermediates['q']

    vK_1 = np.linalg.solve(K.T, v_grids)

    dF, dA = pcm_grad.get_dF_dA(pcmobj.surface)
    dD, dS, dSii = pcm_grad.get_dD_dS(pcmobj.surface, dF, with_D=True, with_S=True)
    DA = D*A

    epsilon = pcmobj.eps

    #de_dF = v0 * -dSii_dF * q
    #de += 0.5*numpy.einsum('i,inx->nx', de_dF, dF)
    # dQ = v^T K^-1 (dR - dK K^-1 R) v
    de = np.zeros([pcmobj.mol.natm,3])
    dD = dD.transpose([2,0,1])
    dS = dS.transpose([2,0,1])
    dSii = dSii.transpose([2,0,1])
    dA = dA.transpose([2,0,1])

    # IEF-PCM and SS(V)PE formally are the same in gradient calculation
    # dR = f_eps/(2*pi) * (dD*A + D*dA),
    # dK = dS - f_eps/(2*pi) * (dD*A*S + D*dA*S + D*A*dS)
    f_epsilon = (epsilon - 1.0)/(epsilon + 1.0)
    fac = f_epsilon/(2.0*np.pi)

    Av = A*v_grids
    de_dR  = 0.5*fac * np.einsum('i,xij,j->ix', vK_1, dD, Av)
    de_dR -= 0.5*fac * np.einsum('i,xij,j->jx', vK_1, dD, Av)
    de_dR  = np.asarray([np.sum(de_dR[p0:p1], axis=0) for p0,p1 in gridslice])

    vK_1_D = vK_1.dot(D)
    vK_1_Dv = vK_1_D * v_grids
    de_dR += 0.5*fac * np.einsum('j,xjn->nx', vK_1_Dv, dA)

    de_dS0  = 0.5*np.einsum('i,xij,j->ix', vK_1, dS, q)
    de_dS0 -= 0.5*np.einsum('i,xij,j->jx', vK_1, dS, q)
    de_dS0  = np.asarray([np.sum(de_dS0[p0:p1], axis=0) for p0,p1 in gridslice])

    vK_1_q = vK_1 * q
    de_dS0 += 0.5*np.einsum('i,xin->nx', vK_1_q, dSii)

    vK_1_DA = np.dot(vK_1, DA)
    de_dS1  = 0.5*np.einsum('i,xij,j->ix', vK_1_DA, dS, q)
    de_dS1 -= 0.5*np.einsum('i,xij,j->jx', vK_1_DA, dS, q)
    de_dS1  = np.asarray([np.sum(de_dS1[p0:p1], axis=0) for p0,p1 in gridslice])

    vK_1_DAq = vK_1_DA*q
    de_dS1 += 0.5*np.einsum('j,xjn->nx', vK_1_DAq, dSii)

    Sq = np.dot(S,q)
    ASq = A*Sq
    de_dD  = 0.5*np.einsum('i,xij,j->ix', vK_1, dD, ASq)
    de_dD -= 0.5*np.einsum('i,xij,j->jx', vK_1, dD, ASq)
    de_dD  = np.asarray([np.sum(de_dD[p0:p1], axis=0) for p0,p1 in gridslice])

    vK_1_D = np.dot(vK_1, D)
    de_dA = 0.5*np.einsum('j,xjn->nx', vK_1_D*Sq, dA)   # 0.5*cupy.einsum('j,xjn,j->nx', vK_1_D, dA, Sq)

    de_dK = de_dS0 - fac * (de_dD + de_dA + de_dS1)
    de += de_dR - de_dK

    t1 = log.timer_debug1('grad solver', *t1)
    return de

def get_cds(smdobj):
    return smd.get_cds_legacy(smdobj)[1]



#!/usr/bin/env python
# Copyright 2020-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
import h5py
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import hf as mol_hf
from pyscf.pbc import tools
from pyscf.pbc.lib import kpts as libkpts
from pyscf.pbc.scf import khf

@lib.with_doc(khf.get_occ.__doc__)
def get_occ(mf, mo_energy_kpts=None, mo_coeff_kpts=None):
    if mo_energy_kpts is None:
        mo_energy_kpts = mf.mo_energy
    cell = mf.cell
    kpts = mf.kpts
    assert isinstance(kpts, libkpts.KPoints)

    nocc = cell.tot_electrons(kpts.nkpts) // 2
    mo_energy_kpts = kpts.transform_mo_energy(mo_energy_kpts)
    mo_energy = np.sort(np.hstack(mo_energy_kpts))
    fermi = mo_energy[nocc-1]
    mo_occ_kpts = []
    for mo_e in mo_energy_kpts:
        mo_occ_kpts.append((mo_e <= fermi).astype(np.double) * 2)

    if nocc < mo_energy.size:
        logger.info(mf, 'HOMO = %.12g  LUMO = %.12g',
                    mo_energy[nocc-1], mo_energy[nocc])
        if mo_energy[nocc-1]+1e-3 > mo_energy[nocc]:
            logger.warn(mf, 'HOMO %.12g == LUMO %.12g',
                        mo_energy[nocc-1], mo_energy[nocc])
    else:
        logger.info(mf, 'HOMO = %.12g', mo_energy[nocc-1])

    if mf.verbose >= logger.DEBUG:
        np.set_printoptions(threshold=len(mo_energy))
        logger.debug(mf, '     k-point                  mo_energy')
        for k,kpt in enumerate(mf.cell.get_scaled_kpts(mf.kpts, kpts_in_ibz=False)):
            logger.debug(mf, '  %2d (%6.3f %6.3f %6.3f)   %s %s',
                         k, kpt[0], kpt[1], kpt[2],
                         np.sort(mo_energy_kpts[k][mo_occ_kpts[k]> 0]),
                         np.sort(mo_energy_kpts[k][mo_occ_kpts[k]==0]))
        np.set_printoptions(threshold=1000)

    mo_occ_kpts = kpts.check_mo_occ_symmetry(mo_occ_kpts)
    return mo_occ_kpts

@lib.with_doc(khf.energy_elec.__doc__)
def energy_elec(mf, dm_kpts=None, h1e_kpts=None, vhf_kpts=None):
    if dm_kpts is None: dm_kpts = mf.make_rdm1()
    if h1e_kpts is None: h1e_kpts = mf.get_hcore()
    if vhf_kpts is None: vhf_kpts = mf.get_veff(mf.cell, dm_kpts)

    kpts_weights = mf.kpts.weights_ibz
    e1 = np.einsum('k,kij,kji', kpts_weights, dm_kpts, h1e_kpts)
    e_coul = np.einsum('k,kij,kji', kpts_weights, dm_kpts, vhf_kpts) * 0.5
    mf.scf_summary['e1'] = e1.real
    mf.scf_summary['e2'] = e_coul.real
    logger.debug(mf, 'E1 = %s  E_coul = %s', e1, e_coul)
    if khf.CHECK_COULOMB_IMAG and abs(e_coul.imag) > mf.cell.precision*10:
        logger.warn(mf, "Coulomb energy has imaginary part %s. "
                    "Coulomb integrals (e-e, e-N) may not converge !",
                    e_coul.imag)
    return (e1+e_coul).real, e_coul.real

@lib.with_doc(khf.get_rho.__doc__)
def get_rho(mf, dm=None, grids=None, kpts=None):
    if isinstance(kpts, np.ndarray):
        return khf.get_rho(mf, dm, grids, kpts)
    if dm is None: dm = mf.make_rdm1()
    if kpts is None: kpts = mf.kpts

    if isinstance(dm[0], np.ndarray) and dm[0].ndim == 3:
        ndm = len(dm[0])
    else:
        ndm = len(dm)
    if ndm != kpts.nkpts_ibz:
        raise RuntimeError("Number of input density matrices does not \
                           match the number of IBZ kpts: %d vs %d."
                           % (ndm, kpts.nkpts_ibz))
    dm = kpts.transform_dm(dm)
    return khf.get_rho(mf, dm, grids, kpts.kpts)

def eig(kmf, h_kpts, s_kpts):
    from pyscf.scf.hf_symm import eig as eig_symm
    cell = kmf.cell
    symm_orb = cell.symm_orb
    irrep_id = cell.irrep_id

    nkpts = len(h_kpts)
    assert len(symm_orb) == nkpts
    eig_kpts = []
    mo_coeff_kpts = []

    for k in range(nkpts):
        e, c = eig_symm(kmf, h_kpts[k], s_kpts[k], symm_orb[k], irrep_id[k])
        eig_kpts.append(e)
        mo_coeff_kpts.append(c)
    return eig_kpts, mo_coeff_kpts

def ksymm_scf_common_init(kmf, cell, kpts, use_ao_symmetry=True):
    kmf._kpts = None
    kmf.use_ao_symmetry = (cell.dimension == 3 and
                           use_ao_symmetry and
                           not kpts.time_reversal and
                           kpts.symmorphic and
                           len(kpts.little_cogroup_ops) > 0)
    if kmf.use_ao_symmetry and cell.symm_orb is None:
        cell._build_symmetry(kpts)
    return kmf


class KsymAdaptedKSCF(khf.KSCF):
    """
    KRHF with k-point symmetry
    """

    _keys = {'use_ao_symmetry'}

    get_occ = get_occ
    get_rho = get_rho
    energy_elec = energy_elec

    def __init__(self, cell, kpts=libkpts.KPoints(),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'),
                 use_ao_symmetry=True):
        ksymm_scf_common_init(self, cell, kpts, use_ao_symmetry)
        khf.KSCF.__init__(self, cell, kpts=kpts, exxdiv=exxdiv)

    @property
    def kpts(self):
        if 'kpts' in self.__dict__:
            # To handle the attribute kpt loaded from chkfile
            kpts_ibz = self.__dict__.pop('kpts')
            if len(kpts_ibz) != self._kpts.nkpts_ibz:
                raise RuntimeError("chkfile is not consistent with the current system.")
        return self._kpts

    @kpts.setter
    def kpts(self, kpts):
        if isinstance(kpts, np.ndarray):
            logger.warn(self, "Input kpts is ndarray, building kpts object without symmetry.")
            kpts = libkpts.make_kpts(self.cell, kpts=kpts)
        elif not isinstance(kpts, libkpts.KPoints):
            raise TypeError("Input kpts have wrong type: %s" % type(kpts))
        kpts_bz = kpts.kpts
        self.with_df.kpts = np.reshape(kpts_bz, (-1,3))
        self._kpts = kpts

    @property
    def kmesh(self):
        from pyscf.pbc.tools.k2gamma import kpts_to_kmesh
        kpts_bz = self._kpts.kpts
        kmesh = kpts_to_kmesh(kpts_bz)
        if len(kpts_bz) != np.prod(kmesh):
            logger.WARN(self, 'K-points specified in %s are not Monkhorst-Pack %s grids',
                        self, kmesh)
        return kmesh

    @kmesh.setter
    def kmesh(self, x):
        self.kpts = self.cell.make_kpts(x)

    def dump_flags(self, verbose=None):
        mol_hf.SCF.dump_flags(self, verbose)
        logger.info(self, '\n')
        logger.info(self, '******** PBC SCF flags ********')
        logger.info(self, 'N kpts (BZ) = %d', self.kpts.nkpts)
        logger.debug(self, 'kpts (BZ) = %s', self.kpts.kpts)
        logger.debug(self, 'kpts weights (BZ) = %s', self.kpts.weights)
        logger.info(self, 'N kpts (IBZ) = %d', self.kpts.nkpts_ibz)
        logger.debug(self, 'kpts (IBZ) = %s', self.kpts.kpts_ibz)
        logger.debug(self, 'kpts weights (IBZ) = %s', self.kpts.weights_ibz)
        logger.info(self, 'Exchange divergence treatment (exxdiv) = %s', self.exxdiv)
        cell = self.cell
        if ((cell.dimension >= 2 and cell.low_dim_ft_type != 'inf_vacuum') and
            isinstance(self.exxdiv, str) and self.exxdiv.lower() == 'ewald'):
            madelung = tools.pbc.madelung(cell, [self.kpts.kpts])
            logger.info(self, '    madelung (= occupied orbital energy shift) = %s', madelung)
            # FIXME: consider the fractional num_electron or not? This maybe
            # relates to the charged system.
            nelectron = float(self.cell.tot_electrons(self.kpts.nkpts)) / self.kpts.nkpts
            logger.info(self, '    Total energy shift due to Ewald probe charge'
                        ' = -1/2 * Nelec*madelung = %.12g',
                        madelung*nelectron * -.5)
        logger.info(self, 'DF object = %s', self.with_df)
        if not getattr(self.with_df, 'build', None):
            self.with_df.dump_flags(verbose)
        return self

    @lib.with_doc(khf.get_ovlp.__doc__)
    def get_ovlp(self, cell=None, kpts=None):
        if isinstance(kpts, np.ndarray):
            return khf.KSCF.get_ovlp(self, cell, kpts)
        if kpts is None: kpts = self.kpts
        return khf.KSCF.get_ovlp(self, cell, kpts.kpts_ibz)

    @lib.with_doc(khf.get_hcore.__doc__)
    def get_hcore(self, cell=None, kpts=None):
        if isinstance(kpts, np.ndarray):
            return khf.KSCF.get_hcore(self, cell, kpts)
        if kpts is None: kpts = self.kpts
        return khf.KSCF.get_hcore(self, cell, kpts.kpts_ibz)

    @lib.with_doc(khf.get_jk.__doc__)
    def get_jk(self, cell=None, dm_kpts=None, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, **kwargs):
        if isinstance(kpts, np.ndarray):
            return super().get_jk(cell, dm_kpts, hermi, kpts, kpts_band,
                                                       with_j, with_k, omega, **kwargs)
        if cell is None: cell = self.cell
        if kpts is None: kpts = self.kpts
        if dm_kpts is None: dm_kpts = self.make_rdm1()
        #get dms for each kpt in BZ
        if isinstance(dm_kpts[0], np.ndarray) and dm_kpts[0].ndim == 3:
            ndm = len(dm_kpts[0])
        else:
            ndm = len(dm_kpts)
        if ndm != kpts.nkpts_ibz:
            raise RuntimeError("Number of input density matrices does not \
                               match the number of IBZ kpts: %d vs %d."
                               % (ndm, kpts.nkpts_ibz))
        dm_kpts = kpts.transform_dm(dm_kpts)
        if kpts_band is None: kpts_band = kpts.kpts_ibz
        cpu0 = (logger.process_clock(), logger.perf_counter())
        if self.rsjk:
            raise NotImplementedError('rsjk with k-points symmetry')
        else:
            vj, vk = self.with_df.get_jk(dm_kpts, hermi, kpts.kpts, kpts_band,
                                         with_j, with_k, omega, exxdiv=self.exxdiv)
        logger.timer(self, 'vj and vk', *cpu0)
        return vj, vk

    def init_guess_by_chkfile(self, chk=None, project=None, kpts=None):
        if isinstance(kpts, np.ndarray):
            return super().init_guess_by_chkfile(chk, project, kpts)
        if kpts is None: kpts = self.kpts
        return super().init_guess_by_chkfile(chk, project, kpts.kpts_ibz)

    def dump_chk(self, envs):
        if self.chkfile:
            mol_hf.SCF.dump_chk(self, envs)
            with lib.H5FileWrap(self.chkfile, 'a') as fh5:
                fh5['scf/kpts'] = self.kpts.kpts_ibz #FIXME Shall we rebuild kpts? If so, more info is needed.
        return self

    def eig(self, h_kpts, s_kpts):
        if self.use_ao_symmetry:
            return eig(self, h_kpts, s_kpts)
        else:
            return khf.KSCF.eig(self, h_kpts, s_kpts)

    def get_orbsym(self, mo_coeff=None, s=None):
        if not self.use_ao_symmetry:
            raise RuntimeError("AO symmetry not initiated")
        from pyscf.scf.hf_symm import get_orbsym
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if s is None:
            s = self.get_ovlp()

        cell = self.cell
        symm_orb = cell.symm_orb
        irrep_id = cell.irrep_id
        orbsym = []
        for k in range(len(mo_coeff)):
            orbsym_k = np.asarray(get_orbsym(cell, mo_coeff[k], s=s[k],
                                             symm_orb=symm_orb[k], irrep_id=irrep_id[k]))
            orbsym.append(orbsym_k)
        return orbsym

    orbsym = property(get_orbsym)

    def _finalize(self):
        khf.KSCF._finalize(self)
        if not self.use_ao_symmetry:
            return self

        orbsym = self.get_orbsym()
        for k, mo_e in enumerate(self.mo_energy):
            idx = np.argsort(mo_e.round(9), kind='stable')
            self.mo_energy[k] = self.mo_energy[k][idx]
            self.mo_occ[k] = self.mo_occ[k][idx]
            self.mo_coeff[k] = lib.tag_array(self.mo_coeff[k][:,idx], orbsym=orbsym[k][idx])

        self.dump_chk({'e_tot': self.e_tot, 'mo_energy': self.mo_energy,
                       'mo_coeff': self.mo_coeff, 'mo_occ': self.mo_occ})
        return self

    def to_khf(self):
        '''transform to non-symmetry object
        '''
        from pyscf.pbc.scf import kuhf_ksymm, kghf_ksymm
        from pyscf.pbc.scf import khf, kuhf, kghf
        from pyscf.pbc.dft import krks, krks_ksymm, kuks, kuks_ksymm
        from pyscf.scf import addons as mol_addons

        def update_mo_(mf, mf1):
            kpts = mf.kpts
            if mf.mo_energy is not None:
                mo_energy = kpts.transform_mo_energy(mf.mo_energy)
                mo_occ = kpts.transform_mo_occ(mf.mo_occ)

                if isinstance(mf, kghf_ksymm.KGHF):
                    mo_coeff = np.asarray(mf.mo_coeff)
                    nao = mo_coeff.shape[1] // 2
                    mo_coeff_alpha = kpts.transform_mo_coeff(mo_coeff[:,:nao])
                    mo_coeff_beta = kpts.transform_mo_coeff(mo_coeff[:,nao:])
                    mo_coeff = []
                    for k in range(len(mo_coeff_alpha)):
                        mo_coeff.append(np.vstack((mo_coeff_alpha[k], mo_coeff_beta[k])))
                    mo_coeff = np.asarray(mo_coeff)
                else:
                    mo_coeff = kpts.transform_mo_coeff(mf.mo_coeff)

                mf1.mo_coeff = mo_coeff
                mf1.mo_occ = mo_occ
                mf1.mo_energy = mo_energy
            return mf1

        known_cls = {KsymAdaptedKRHF : khf.KRHF,
                     kuhf_ksymm.KUHF : kuhf.KUHF,
                     kghf_ksymm.KGHF : kghf.KGHF,
                     krks_ksymm.KRKS : krks.KRKS,
                     kuks_ksymm.KUKS : kuks.KUKS}

        out = mol_addons._object_without_soscf(self, known_cls, False)
        out.__dict__.pop('kpts', None)
        return update_mo_(self, out)

    def sfx2c1e(self):
        raise NotImplementedError
    x2c = x2c1e = sfx2c1e


class KsymAdaptedKRHF(KsymAdaptedKSCF, khf.KRHF):

    to_ks = khf.KRHF.to_ks
    convert_from_ = khf.KRHF.convert_from_

    def get_init_guess(self, cell=None, key='minao', s1e=None):
        if s1e is None:
            s1e = self.get_ovlp(cell)
        dm_kpts = mol_hf.SCF.get_init_guess(self, cell, key)
        if dm_kpts.ndim == 2:
            dm_kpts = np.asarray([dm_kpts]*self.kpts.nkpts_ibz)
        elif len(dm_kpts) != self.kpts.nkpts_ibz:
            dm_kpts = dm_kpts[self.kpts.ibz2bz]

        ne = lib.einsum('k,kij,kji', self.kpts.weights_ibz, dm_kpts, s1e).real
        nkpts = self.kpts.nkpts
        ne *= nkpts
        nelectron = float(self.cell.tot_electrons(nkpts))
        if abs(ne - nelectron) > 0.01*nkpts:
            logger.debug(self, 'Big error detected in the electron number '
                         'of initial guess density matrix (Ne/cell = %g)!\n'
                         '  This can cause huge error in Fock matrix and '
                         'lead to instability in SCF for low-dimensional '
                         'systems.\n  DM is normalized wrt the number '
                         'of electrons %s', ne/nkpts, nelectron/nkpts)
            dm_kpts *= (nelectron / ne).reshape(-1,1,1)
        return dm_kpts

KRHF = KsymAdaptedKRHF



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

import ctypes
import math
import numpy
from pyscf import lib

libfci = lib.load_library('libfci')

def make_strings(orb_list, nelec):
    '''Generate string from the given orbital list.

    Returns:
        list of int64.  One element represents one string in binary format.
        The binary format takes the convention that the one bit stands for one
        orbital, bit-1 means occupied and bit-0 means unoccupied.  The lowest
        (right-most) bit corresponds to the lowest orbital in the orb_list.

    Examples:

    >>> [bin(x) for x in make_strings((0,1,2,3),2)]
    [0b11, 0b101, 0b110, 0b1001, 0b1010, 0b1100]
    >>> [bin(x) for x in make_strings((3,1,0,2),2)]
    [0b1010, 0b1001, 0b11, 0b1100, 0b110, 0b101]
    '''
    orb_list = list(orb_list)
    if len(orb_list) >= 64:
        return gen_occslst(orb_list, nelec)

    assert (nelec >= 0)
    if nelec == 0:
        return numpy.asarray([0], dtype=numpy.int64)
    elif nelec > len(orb_list):
        return numpy.asarray([], dtype=numpy.int64)
    def gen_str_iter(orb_list, nelec):
        if nelec == 1:
            res = [(1 << i) for i in orb_list]
        elif nelec >= len(orb_list):
            n = 0
            for i in orb_list:
                n = n | (1 << i)
            res = [n]
        else:
            restorb = orb_list[:-1]
            thisorb = 1 << orb_list[-1]
            res = gen_str_iter(restorb, nelec)
            for n in gen_str_iter(restorb, nelec-1):
                res.append(n | thisorb)
        return res
    strings = gen_str_iter(orb_list, nelec)
    assert (strings.__len__() == num_strings(len(orb_list),nelec))
    return numpy.asarray(strings, dtype=numpy.int64)
gen_strings4orblist = make_strings

def gen_occslst(orb_list, nelec):
    '''Generate occupied orbital list for each string.

    Returns:
        List of lists of int32. Each inner list has length equal to the number of
        electrons, and contains the occupied orbitals in the corresponding string.

    Example:

        >>> [bin(x) for x in make_strings((0, 1, 2, 3), 2)]
        ['0b11', '0b101', '0b110', '0b1001', '0b1010', '0b1100']
        >>> gen_occslst((0, 1, 2, 3), 2)
        OIndexList([[0, 1],
                    [0, 2],
                    [1, 2],
                    [0, 3],
                    [1, 3],
                    [2, 3]], dtype=int32)
    '''
    orb_list = list(orb_list)
    assert (nelec >= 0)
    if nelec == 0:
        return numpy.zeros((1,nelec), dtype=numpy.int32)
    elif nelec > len(orb_list):
        return numpy.zeros((0,nelec), dtype=numpy.int32)
    def gen_occs_iter(orb_list, nelec):
        if nelec == 1:
            res = [[i] for i in orb_list]
        elif nelec >= len(orb_list):
            res = [orb_list]
        else:
            restorb = orb_list[:-1]
            thisorb = orb_list[-1]
            res = gen_occs_iter(restorb, nelec)
            for n in gen_occs_iter(restorb, nelec-1):
                res.append(n + [thisorb])
        return res
    occslst = gen_occs_iter(orb_list, nelec)
    return numpy.asarray(occslst, dtype=numpy.int32).view(OIndexList)
# Add this symbol for backward compatibility. Should remove in the future.
_gen_occslst = gen_occslst

def _strs2occslst(strs, norb):
    na = len(strs)
    one_particle_strs = numpy.asarray([1 << i for i in range(norb)])
    occ_masks = (strs.reshape(-1,1) & one_particle_strs) != 0
    occslst = numpy.where(occ_masks)[1].reshape(na,-1)
    return numpy.asarray(occslst, dtype=numpy.int32).view(OIndexList)

def _occslst2strs(occslst):
    assert isinstance(occslst[0], OIndexList)
    occslst = numpy.asarray(occslst)
    na, nelec = occslst.shape
    strs = numpy.zeros(na, dtype=numpy.int64)
    for i in range(nelec):
        strs ^= 1 << occslst[:,i]
    return strs

class OIndexList(numpy.ndarray):
    pass

num_strings = lib.comb

def gen_linkstr_index_o1(orb_list, nelec, strs=None, tril=False):
    '''Look up table, for the strings relationship in terms of a
    creation-annihilating operator pair.

    Similar to gen_linkstr_index. The only difference is the input argument
    strs, which needs to be a list of OIndexList in this function.
    '''
    if nelec == 0:
        return numpy.zeros((0,0,4), dtype=numpy.int32)

    if strs is None:
        strs = gen_occslst(orb_list, nelec)
    occslst = strs

    orb_list = numpy.asarray(orb_list)
    norb = len(orb_list)
    assert (numpy.all(numpy.arange(norb) == orb_list))

    strdic = {tuple(s): i for i,s in enumerate(occslst)}
    nvir = norb - nelec
    def propgate1e(str0):
        addr0 = strdic[tuple(str0)]
        tab = numpy.empty((nelec,4), dtype=numpy.int32)
        tab[:,0] = tab[:,1] = str0
        tab[:,2] = addr0
        tab[:,3] = 1
        linktab = [tab]

        virmask = numpy.ones(norb, dtype=bool)
        virmask[str0] = False
        vir = orb_list[virmask]
        str0 = numpy.asarray(str0)
        # where to put vir-orb, ie how many occ-orb in the left
        where_vir = numpy.sum(str0.reshape(-1,1) < vir, axis=0)
        parity_occ_orb = 1  # parity for annihilating occupied orbital
        for n,i in enumerate(str0):  # loop over all occupied orbitals
            # o,v which index is bigger, to determine whether to annihilate occ-orb first
            reorder_to_ov = vir > i
            str1s = numpy.empty((nvir,nelec), dtype=int)
            str1s[:] = str0
            str1s[:,n] = vir
            str1s.sort(axis=1)
            addr = [strdic[tuple(s)] for s in str1s]
            parity = (where_vir + reorder_to_ov + 1) % 2 #? +1 so that even parity has +1, odd parity = 0
            parity[parity == 0] = -1
            parity *= parity_occ_orb
            tab = numpy.empty((nvir,4), dtype=numpy.int32)
            tab[:,0] = vir
            tab[:,1] = i
            tab[:,2] = addr
            tab[:,3] = parity
            linktab.append(tab)
            parity_occ_orb *= -1
        return numpy.vstack(linktab)

    lidx = [propgate1e(s) for s in occslst]
    lidx = numpy.asarray(lidx, dtype=numpy.int32)
    if tril:
        lidx = reform_linkstr_index(lidx)
    return lidx

# return [cre, des, target_address, parity]
def gen_linkstr_index(orb_list, nocc, strs=None, tril=False):
    '''Look up table, for the strings relationship in terms of a
    creation-annihilating operator pair.

    For given string str0, index[str0] is (nocc+nocc*nvir) x 4 array.
    The first nocc rows [i(:occ),i(:occ),str0,sign] are occupied-occupied
    excitations, which do not change the string. The next nocc*nvir rows
    [a(:vir),i(:occ),str1,sign] are occupied-virtual excitations, starting from
    str0, annihilating i, creating a, to get str1.
    '''
    if strs is None:
        strs = make_strings(orb_list, nocc)

    if isinstance(strs, OIndexList):
        return gen_linkstr_index_o1(orb_list, nocc, strs, tril)

    strs = numpy.array(strs, dtype=numpy.int64)
    assert (all(strs[:-1] < strs[1:]))
    norb = len(orb_list)
    nvir = norb - nocc
    na = strs.shape[0]
    link_index = numpy.empty((na,nocc*nvir+nocc,4), dtype=numpy.int32)
    libfci.FCIlinkstr_index(link_index.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb), ctypes.c_int(na),
                            ctypes.c_int(nocc),
                            strs.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(tril))
    return link_index

def reform_linkstr_index(link_index):
    '''Compress the (a, i) pair index in linkstr_index to a lower triangular
    index. The compressed indices can match the 4-fold symmetry of integrals.
    '''
    #for k, tab in enumerate(link_index):
    #    for j, (a, i, str1, sign) in enumerate(tab):
    #        if a > i:
    #            ai = a*(a+1)//2+i
    #        else:
    #            ai = i*(i+1)//2+a
    #        link_new[k,j] = (ai,0,str1,sign)
    link_new = link_index.copy()
    a = link_index[:,:,0]
    i = link_index[:,:,1]
    link_new[:,:,0] = numpy.maximum(a*(a+1)//2+i, i*(i+1)//2+a)
    link_new[:,:,1] = 0
    return link_new

def gen_linkstr_index_trilidx(orb_list, nocc, strs=None):
    r'''Generate linkstr_index with the assumption that :math:`p^+ q|0\rangle`
    where :math:`p > q`.
    So the resultant link_index has the structure ``[pq, *, str1, sign]``.
    It is identical to a call to ``reform_linkstr_index(gen_linkstr_index(...))``.
    '''
    return gen_linkstr_index(orb_list, nocc, strs, True)

def gen_cre_str_index(orb_list, nelec):
    '''linkstr_index to map between N electron string to N+1 electron string.
    It maps the given string to the address of the string which is generated by
    the creation operator.

    For given string str0, index[str0] is nvir x 4 array.  Each entry
    [i(cre),--,str1,sign] means starting from str0, creating i, to get str1.

    Returns [[cre, -, target_address, parity], ...]
    '''
    norb = len(orb_list)
    assert nelec < norb
    strs = make_strings(orb_list, nelec)
    if isinstance(strs, OIndexList):
        raise NotImplementedError('System with 64 orbitals or more')

    strs = numpy.array(strs, dtype=numpy.int64)
    na = strs.shape[0]
    link_index = numpy.empty((len(strs),norb-nelec,4), dtype=numpy.int32)
    libfci.FCIcre_str_index(link_index.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb), ctypes.c_int(na),
                            ctypes.c_int(nelec),
                            strs.ctypes.data_as(ctypes.c_void_p))
    return link_index

def gen_des_str_index(orb_list, nelec):
    '''linkstr_index to map between N electron string to N-1 electron string.
    It maps the given string to the address of the string which is generated by
    the annihilation operator.

    For given string str0, index[str0] is nvir x 4 array.  Each entry
    [--,i(des),str1,sign] means starting from str0, annihilating i, to get str1.

    Returns [[-, des, target_address, parity], ...]
    '''
    assert nelec > 0
    strs = make_strings(orb_list, nelec)
    if isinstance(strs, OIndexList):
        raise NotImplementedError('System with 64 orbitals or more')

    strs = numpy.array(strs, dtype=numpy.int64)
    norb = len(orb_list)
    na = strs.shape[0]
    link_index = numpy.empty((len(strs),nelec,4), dtype=numpy.int32)
    libfci.FCIdes_str_index(link_index.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb), ctypes.c_int(na),
                            ctypes.c_int(nelec),
                            strs.ctypes.data_as(ctypes.c_void_p))
    return link_index


# Determine the sign of  p^+ q |string0>
def cre_des_sign(p, q, string0):
    if p == q:
        return 1
    else:
        if (string0 & (1 << p)) or (not (string0 & (1 << q))):
            return 0
        elif p > q:
            mask = (1 << p) - (1 << (q+1))
        else:
            mask = (1 << q) - (1 << (p+1))
        return (-1) ** bin(string0 & mask).count('1')

# Determine the sign of  p^+ |string0>
def cre_sign(p, string0):
    if (string0 & (1 << p)):
        return 0
    else:
        return (-1) ** bin(string0 >> (p+1)).count('1')

# Determine the sign of  p |string0>
def des_sign(p, string0):
    if (not (string0 & (1 << p))):
        return 0
    else:
        return (-1) ** bin(string0 >> (p+1)).count('1')

# Determine the sign of  string1 = p^+ q |string0>
def parity(string0, string1):
    #sys.stderr.write('Function cistring.parity is deprecated\n')
    ss = string1 - string0
    def count_bit1(n):
        # see Hamming weight problem and K&R C program
        return bin(n).count('1')
    if ss > 0:
        # string1&ss gives the number of 1s between two strings
        return (-1) ** (count_bit1(string1 & ss))
    elif ss == 0:
        return 1
    else:
        return (-1) ** (count_bit1(string0 & (-ss)))

def addr2str(norb, nelec, addr):
    '''Convert CI determinant address to string'''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')
    max_addr = num_strings(norb, nelec)
    assert max_addr > addr

    if max_addr < 2**31:
        return addrs2str(norb, nelec, [addr])[0]
    else:
        return _addr2str(norb, nelec, addr)

def _addr2str(norb, nelec, addr):
    if addr == 0 or nelec == norb or nelec == 0:
        return (1 << nelec) - 1   # ..0011..11

    for i in reversed(range(norb)):
        addrcum = num_strings(i, nelec)
        if addrcum <= addr:
            return (1 << i) | _addr2str(i, nelec-1, addr-addrcum)

def addrs2str(norb, nelec, addrs):
    '''Convert a list of CI determinant address to string'''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')
    if num_strings(norb, nelec) >= 2**31:
        raise NotImplementedError('Large address')

    addrs = numpy.asarray(addrs, dtype=numpy.int32)
    assert (all(num_strings(norb, nelec) > addrs))
    count = addrs.size
    strs = numpy.empty(count, dtype=numpy.int64)
    libfci.FCIaddrs2str(strs.ctypes.data_as(ctypes.c_void_p),
                        addrs.ctypes.data_as(ctypes.c_void_p),
                        ctypes.c_int(count),
                        ctypes.c_int(norb), ctypes.c_int(nelec))
    return strs

def str2addr(norb, nelec, string):
    '''Convert string to CI determinant address'''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')

    if isinstance(string, str):
        assert (string.count('1') == nelec)
        string = int(string, 2)
    else:
        assert (bin(string).count('1') == nelec)

    if num_strings(norb, nelec) < 2**31:
        libfci.FCIstr2addr.restype = ctypes.c_int
        return libfci.FCIstr2addr(ctypes.c_int(norb), ctypes.c_int(nelec),
                                  ctypes.c_ulonglong(string))
    return _str2addr(norb, nelec, string)

def _str2addr(norb, nelec, string):
    if norb <= nelec or nelec == 0:
        return 0
    addr = 0
    for orbital_id in reversed(range(norb)):
        if (1 << orbital_id) & string:
            addr += num_strings(orbital_id, nelec)
            nelec -= 1
    return addr

def strs2addr(norb, nelec, strings):
    '''Convert a list of string to CI determinant address'''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')
    if num_strings(norb, nelec) >= 2**31:
        raise NotImplementedError('Large address')

    strings = numpy.asarray(strings, dtype=numpy.int64)
    count = strings.size
    addrs = numpy.empty(count, dtype=numpy.int32)
    libfci.FCIstrs2addr(addrs.ctypes.data_as(ctypes.c_void_p),
                        strings.ctypes.data_as(ctypes.c_void_p),
                        ctypes.c_int(count),
                        ctypes.c_int(norb), ctypes.c_int(nelec))
    return addrs

def sub_addrs(norb, nelec, orbital_indices, sub_nelec=0):
    '''The addresses of the determinants which include the specified orbital
    indices. The size of the returned addresses is equal to the number of
    determinants of (norb, nelec) system.
    '''
    assert norb < 64
    if sub_nelec == 0:
        strs = make_strings(orbital_indices, nelec)
        return strs2addr(norb, nelec, strs)
    else:
        strs = make_strings(range(norb), nelec)
        counts = numpy.zeros(len(strs), dtype=int)
        for i in orbital_indices:
            counts += (strs & (1 << i)) != 0
        sub_strs = strs[counts == sub_nelec]
        return strs2addr(norb, nelec, sub_strs)

def tn_strs(norb, nelec, n):
    '''Generate strings for Tn amplitudes.  Eg n=1 (T1) has nvir*nocc strings,
    n=2 (T2) has nvir*(nvir-1)/2 * nocc*(nocc-1)/2 strings.
    '''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')

    if nelec < n or norb-nelec < n:
        return numpy.zeros(0, dtype=int)
    occs_allow = numpy.asarray(make_strings(range(nelec), n)[::-1])
    virs_allow = numpy.asarray(make_strings(range(nelec,norb), n))
    hf_str = int('1'*nelec, 2)
    tns = (hf_str | virs_allow.reshape(-1,1)) ^ occs_allow
    return tns.ravel()

if __name__ == '__main__':
    print([bin(i) for i in make_strings(range(2,5), 2)])
    print(make_strings(range(4), 2))
    #print(gen_linkstr_index(range(6), 3))
#    index = gen_linkstr_index(range(8), 4)
#    idx16 = index[:16]
#    print(idx16[:,:,2])



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Restricted open-shell Kohn-Sham for periodic systems with k-point sampling
'''

import numpy as np
from pyscf import lib
from pyscf.pbc.scf import krohf
from pyscf.pbc.dft import rks
from pyscf.pbc.dft import kuks
from pyscf.pbc.dft.kuks import energy_elec
from pyscf import __config__


@lib.with_doc(kuks.get_veff.__doc__)
def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpts=None, kpts_band=None):
    if getattr(dm, 'mo_coeff', None) is not None:
        mo_coeff = dm.mo_coeff
        mo_occ_a = [(x > 0).astype(np.double) for x in dm.mo_occ]
        mo_occ_b = [(x ==2).astype(np.double) for x in dm.mo_occ]
        dm = lib.tag_array(dm, mo_coeff=(mo_coeff,mo_coeff),
                           mo_occ=(mo_occ_a,mo_occ_b))
    return kuks.get_veff(ks, cell, dm, dm_last, vhf_last, hermi, kpts, kpts_band)


class KROKS(rks.KohnShamDFT, krohf.KROHF):
    '''RKS class adapted for PBCs with k-point sampling.
    '''

    get_veff = get_veff
    energy_elec = energy_elec
    get_rho = kuks.get_rho

    def __init__(self, cell, kpts=np.zeros((1,3)), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald')):
        krohf.KROHF.__init__(self, cell, kpts, exxdiv=exxdiv)
        rks.KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        krohf.KROHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    def to_hf(self):
        '''Convert to KROHF object.'''
        from pyscf.pbc import scf, df
        out = self._transfer_attrs_(scf.KROHF(self.cell, self.kpts))
        # Pure functionals only construct J-type integrals. Enable all integrals for KHF.
        if (not self._numint.libxc.is_hybrid_xc(self.xc) and
            len(self.kpts) > 1 and getattr(self.with_df, '_j_only', False)):
            out.with_df._j_only = False
            out.with_df.reset()
        return out

    multigrid_numint = kuks.KUKS.multigrid_numint

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import ctypes
import numpy as np
from pyscf import lib
from pyscf.lib import logger

libpbc = lib.load_library('libpbc')

class _CNeighborPair(ctypes.Structure):
    __slots__ = []
    _fields_ = [("nimgs", ctypes.c_int),
                ("Ls_list", ctypes.POINTER(ctypes.c_int)),
                ("q_cond", ctypes.POINTER(ctypes.c_double)),
                ("center", ctypes.POINTER(ctypes.c_double))]


class _CNeighborList(ctypes.Structure):
    __slots__ = []
    _fields_ = [("nish", ctypes.c_int),
                ("njsh", ctypes.c_int),
                ("nimgs", ctypes.c_int),
                ("pairs", ctypes.POINTER(ctypes.POINTER(_CNeighborPair)))]


class _CNeighborListOpt(ctypes.Structure):
    __slots__ = []
    _fields_ = [("nl", ctypes.POINTER(_CNeighborList)),
                ('fprescreen', ctypes.c_void_p)]


def build_neighbor_list_for_shlpairs(cell, cell1=None, Ls=None,
                                     ish_rcut=None, jsh_rcut=None, hermi=0,
                                     precision=None):
    '''
    Build the neighbor list of shell pairs for periodic calculations.

    Arguments:
        cell : :class:`pbc.gto.cell.Cell`
            The :class:`Cell` instance for the bra basis functions.
        cell1 : :class:`pbc.gto.cell.Cell`, optional
            The :class:`Cell` instance for the ket basis functions.
            If not given, both bra and ket basis functions come from cell.
        Ls : ``(*,3)`` array, optional
            The cartesian coordinates of the periodic images.
            Default is calculated by :func:`cell.get_lattice_Ls`.
        ish_rcut : (nish,) array, optional
            The cutoff radii of the shells for bra basis functions.
        jsh_rcut : (njsh,) array, optional
            The cutoff radii of the shells for ket basis functions.
        hermi : int, optional
            If :math:`hermi=1`, the task list is built only for
            the upper triangle of the matrix. Default is 0.
        precision : float, optional
            The integral precision. Default is :attr:`cell.precision`.
            If both ``ish_rcut`` and ``jsh_rcut`` are given,
            ``precision`` will be ignored.

    Returns: :class:`ctypes.POINTER`
        The C pointer of the :class:`NeighborList` structure.
    '''
    if cell1 is None:
        cell1 = cell

    if hermi == 1 and cell1 is not cell:
        logger.warn(cell,
                    "Set hermi=0 because cell and cell1 are not the same.")
        hermi = 0

    ish_atm = np.asarray(cell._atm, order='C', dtype=np.int32)
    ish_bas = np.asarray(cell._bas, order='C', dtype=np.int32)
    ish_env = np.asarray(cell._env, order='C', dtype=float)
    nish = len(ish_bas)
    if ish_rcut is None:
        ish_rcut = cell.rcut_by_shells(precision=precision)
    assert nish == len(ish_rcut)

    if cell1 is cell:
        jsh_atm = ish_atm
        jsh_bas = ish_bas
        jsh_env = ish_env
        if jsh_rcut is None:
            jsh_rcut = ish_rcut
    else:
        jsh_atm = np.asarray(cell1._atm, order='C', dtype=np.int32)
        jsh_bas = np.asarray(cell1._bas, order='C', dtype=np.int32)
        jsh_env = np.asarray(cell1._env, order='C', dtype=float)
        if jsh_rcut is None:
            jsh_rcut = cell1.rcut_by_shells(precision=precision)
    njsh = len(jsh_bas)
    assert njsh == len(jsh_rcut)

    if Ls is None:
        rcut = ish_rcut.max(initial=0) + jsh_rcut.max(initial=0)
        Ls = cell.get_lattice_Ls(rcut=rcut)
    Ls = np.asarray(Ls, order='C', dtype=float)
    nimgs = len(Ls)

    nl = ctypes.POINTER(_CNeighborList)()
    libpbc.build_neighbor_list(
        ctypes.byref(nl),
        ish_atm.ctypes.data_as(ctypes.c_void_p),
        ish_bas.ctypes.data_as(ctypes.c_void_p),
        ish_env.ctypes.data_as(ctypes.c_void_p),
        ish_rcut.ctypes.data_as(ctypes.c_void_p),
        jsh_atm.ctypes.data_as(ctypes.c_void_p),
        jsh_bas.ctypes.data_as(ctypes.c_void_p),
        jsh_env.ctypes.data_as(ctypes.c_void_p),
        jsh_rcut.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nish),
        ctypes.c_int(njsh),
        Ls.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nimgs),
        ctypes.c_int(hermi),
    )
    return nl

def free_neighbor_list(nl):
    libpbc.del_neighbor_list(ctypes.byref(nl))

def neighbor_list_to_ndarray(cell, cell1, nl):
    '''
    Returns:
        Ls_list: (nLtot,) ndarray
            indices of Ls
        Ls_idx: (2 x nish x njsh,) ndarray
            starting and ending indices in Ls_list
    '''
    nish = cell.nbas
    njsh = cell1.nbas
    Ls_list = []
    Ls_idx = []
    nLtot = 0
    for i in range(nish):
        for j in range(njsh):
            pair = nl.contents.pairs[i*njsh+j]
            nL = pair.contents.nimgs
            nLtot += nL
            for iL in range(nL):
                idx = pair.contents.Ls_list[iL]
                Ls_list.append(idx)
            if nL > 0:
                Ls_idx.extend([nLtot-nL, nLtot])
            else:
                Ls_idx.extend([-1,-1])
    return np.asarray(Ls_list), np.asarray(Ls_idx)


class NeighborListOpt():
    def __init__(self, cell):
        self.cell = cell
        self.nl = None
        self._this = ctypes.POINTER(_CNeighborListOpt)()
        libpbc.NLOpt_init(ctypes.byref(self._this))

    def build(self, cell=None, cell1=None, Ls=None,
              ish_rcut=None, jsh_rcut=None,
              hermi=0, precision=None,
              set_nl=True, set_optimizer=True):
        if cell is None:
            cell = self.cell

        if (set_nl or set_optimizer) and self.nl is None:
            self.nl = build_neighbor_list_for_shlpairs(
                            cell, cell1=cell1, Ls=Ls,
                            ish_rcut=ish_rcut, jsh_rcut=jsh_rcut,
                            hermi=hermi, precision=precision)
            libpbc.NLOpt_set_nl(self._this, self.nl)

        if set_optimizer:
            libpbc.NLOpt_set_optimizer(self._this)

    def reset(self, free_nl=True):
        if self.nl is not None and free_nl:
            free_neighbor_list(self.nl)
        self.nl = None
        libpbc.NLOpt_reset(self._this)

    def __del__(self):
        self.reset()
        try:
            libpbc.NLOpt_del(ctypes.byref(self._this))
        except AttributeError:
            pass



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import time
import numpy as np
from scipy.linalg import block_diag
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger, einsum
from pyscf.lib.parameters import LARGE_DENOM
from pyscf.pbc.mp import kmp2

WITH_T2 = getattr(__config__, 'mp_mp2_with_t2', False)

def kernel(mp, mo_energy, mo_coeff, verbose=logger.NOTE, with_t2=WITH_T2):
    if with_t2:
        return kernel_with_t2(mp, mo_energy, mo_coeff, verbose, with_t2)
    else:
        t2 = None

    t0 = (logger.process_clock(), logger.perf_counter())
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    nkpts = mp.nkpts
    kd = mp.kpts

    eia = np.zeros((nocc,nvir))
    eijab = np.zeros((nocc,nocc,nvir,nvir))

    fao2mo = mp._scf.with_df.ao2mo
    oovv_ij = np.zeros((nkpts,nocc,nocc,nvir,nvir), dtype=mo_coeff[0].dtype)

    mo_e_o = [mo_energy[k][:nocc] for k in range(nkpts)]
    mo_e_v = [mo_energy[k][nocc:] for k in range(nkpts)]

    # Get location of non-zero/padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = kmp2.padding_k_idx(mp, kind="split")

    kijab, weight, k4_bz2ibz = kd.make_k4_ibz(sym='s2')
    _, igroup = np.unique(kijab[:,:2], axis=0, return_index=True)
    igroup = igroup.ravel()
    igroup = list(igroup) + [len(kijab)]

    emp2_ss = emp2_os = 0.
    nao2mo = 0
    icount = 0
    for i in range(len(igroup)-1):
        istart = igroup[i]
        iend = igroup[i+1]
        kab = []
        for j in range(istart, iend):
            a, b = kijab[j][2:]
            kab.append([a, b])
            kab.append([b, a])
        kab = np.unique(np.asarray(kab), axis=0)

        ki = kijab[istart][0]
        kj = kijab[istart][1]
        kpts_i = kd.kpts[ki]
        kpts_j = kd.kpts[kj]
        orbo_i = mo_coeff[ki][:,:nocc]
        orbo_j = mo_coeff[kj][:,:nocc]

        for (ka, kb) in kab:
            kpts_a = kd.kpts[ka]
            kpts_b = kd.kpts[kb]
            orbv_a = mo_coeff[ka][:,nocc:]
            orbv_b = mo_coeff[kb][:,nocc:]
            oovv_ij[ka] = fao2mo((orbo_i,orbv_a,orbo_j,orbv_b),
                                 (kpts_i,kpts_a,kpts_j,kpts_b),
                                 compact=False).reshape(nocc,nvir,nocc,nvir).transpose(0,2,1,3) / nkpts
            nao2mo += 1

        for j in range(istart, iend):
            ka = kijab[j][2]
            kb = kijab[j][3]
            # Remove zero/padded elements from denominator
            eia = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
            n0_ovp_ia = np.ix_(nonzero_opadding[ki], nonzero_vpadding[ka])
            eia[n0_ovp_ia] = (mo_e_o[ki][:,None] - mo_e_v[ka])[n0_ovp_ia]

            ejb = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
            n0_ovp_jb = np.ix_(nonzero_opadding[kj], nonzero_vpadding[kb])
            ejb[n0_ovp_jb] = (mo_e_o[kj][:,None] - mo_e_v[kb])[n0_ovp_jb]

            eijab = lib.direct_sum('ia,jb->ijab',eia,ejb)
            t2_ijab = np.conj(oovv_ij[ka]/eijab)
            idx_ibz = k4_bz2ibz[ki*nkpts**2 + kj*nkpts + ka]
            assert(icount == idx_ibz)
            edi = einsum('ijab,ijab', t2_ijab, oovv_ij[ka]).real * 2
            exi = -einsum('ijab,ijba', t2_ijab, oovv_ij[kb]).real
            emp2_ss += (edi*0.5 + exi) * weight[idx_ibz] * nkpts**3
            emp2_os += edi*0.5 * weight[idx_ibz] * nkpts**3
            icount += 1

    emp2_ss /= nkpts
    emp2_os /= nkpts
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)
    assert(icount == len(kijab))
    logger.debug(mp, "Number of ao2mo transformations performed in KMP2: %d", nao2mo)
    logger.timer(mp, 'KMP2', *t0)
    return emp2, t2

def kernel_with_t2(mp, mo_energy, mo_coeff, verbose=logger.NOTE, with_t2=WITH_T2):
    #we need almost all t2 for computing rdm, so simply use kmp2 without symmetry
    kd = mp.kpts
    mp.kpts = kd.kpts
    emp2, t2 = kmp2.kernel(mp, mo_energy, mo_coeff, verbose, with_t2)
    mp.kpts = kd
    return emp2, t2

@lib.with_doc(kmp2.make_rdm1.__doc__)
def make_rdm1(mp, t2=None, kind="compact"):
    if kind not in ("compact", "padded"):
        raise ValueError("The 'kind' argument should be either 'compact' or 'padded'")
    d_imds = _gamma1_intermediates(mp, t2=t2)
    result = []
    padding_idxs = kmp2.padding_k_idx(mp, kind="joint")
    for (oo, vv), idxs in zip(zip(*d_imds), padding_idxs):
        oo += np.eye(*oo.shape)
        d = block_diag(oo, vv)
        d += d.conj().T
        if kind == "padded":
            result.append(d)
        else:
            result.append(d[np.ix_(idxs, idxs)])
    return result

def make_t2_for_rdm1(mp):
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    nkpts = mp.nkpts
    kd = mp.kpts
    kpts = kd.kpts
    k_ibz = kd.ibz2bz

    mo_coeff, mo_energy = kmp2._add_padding(mp, mp.mo_coeff, mp.mo_energy)
    eia = np.zeros((nocc,nvir))
    eijab = np.zeros((nocc,nocc,nvir,nvir))
    fao2mo = mp._scf.with_df.ao2mo
    mo_e_o = [mo_energy[k][:nocc] for k in range(nkpts)]
    mo_e_v = [mo_energy[k][nocc:] for k in range(nkpts)]
    nonzero_opadding, nonzero_vpadding = kmp2.padding_k_idx(mp, kind="split")
    t2 = np.empty((nkpts,nkpts,nkpts,nocc,nocc,nvir,nvir), dtype=mo_coeff[0].dtype)

    nao2mo = 0
    for ki in range(nkpts):
        orbo_i = mo_coeff[ki][:,:nocc]
        for kj in range(ki, nkpts):
            orbo_j = mo_coeff[kj][:,:nocc]
            for ka in range(nkpts):
                kb = mp.khelper.kconserv[ki, ka, kj]
                if ki in k_ibz or kj in k_ibz or ka in k_ibz or kb in k_ibz:
                    nao2mo +=1
                    orbv_a = mo_coeff[ka][:,nocc:]
                    orbv_b = mo_coeff[kb][:,nocc:]
                    oovv = fao2mo((orbo_i,orbv_a,orbo_j,orbv_b),
                                  (kpts[ki],kpts[ka],kpts[kj],kpts[kb]),
                                  compact=False).reshape(nocc,nvir,nocc,nvir).transpose(0,2,1,3) / nkpts

                    eia = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
                    n0_ovp_ia = np.ix_(nonzero_opadding[ki], nonzero_vpadding[ka])
                    eia[n0_ovp_ia] = (mo_e_o[ki][:,None] - mo_e_v[ka])[n0_ovp_ia]

                    ejb = LARGE_DENOM * np.ones((nocc, nvir), dtype=mo_energy[0].dtype)
                    n0_ovp_jb = np.ix_(nonzero_opadding[kj], nonzero_vpadding[kb])
                    ejb[n0_ovp_jb] = (mo_e_o[kj][:,None] - mo_e_v[kb])[n0_ovp_jb]

                    eijab = lib.direct_sum('ia,jb->ijab',eia,ejb)
                    t2[ki,kj,ka] = np.conj(oovv / eijab)
    logger.debug(mp, "Number of ao2mo transformations performed in KMP2: %d", nao2mo)
    return t2

def _gamma1_intermediates(mp, t2=None):
    if t2 is None:
        t2 = mp.t2
    if t2 is None:
        t2 = make_t2_for_rdm1(mp)
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    kd = mp.kpts
    nkpts = kd.nkpts
    nkpts_ibz = kd.nkpts_ibz
    dtype = t2.dtype

    dm1occ = np.zeros((nkpts_ibz, nocc, nocc), dtype=dtype)
    dm1vir = np.zeros((nkpts_ibz, nvir, nvir), dtype=dtype)

    k_ibz = kd.ibz2bz
    for ki in range(nkpts):
        for kj in range(nkpts):
            for ka in range(nkpts):
                kb = mp.khelper.kconserv[ki, ka, kj]
                t2a = t2[ki,kj,ka]
                t2b = t2[ki,kj,kb]
                if ki > kj:
                    t2a = t2[kj,ki,kb].transpose(1,0,3,2)
                    t2b = t2[kj,ki,ka].transpose(1,0,3,2)
                if kb in k_ibz:
                    dm1vir[kd.bz2ibz[kb]] += einsum('ijax,ijay->yx', t2a.conj(), t2a) * 2 -\
                                             einsum('ijax,ijya->yx', t2a.conj(), t2b)
                if kj in k_ibz:
                    dm1occ[kd.bz2ibz[kj]] += einsum('ixab,iyab->xy', t2a.conj(), t2a) * 2 -\
                                             einsum('ixab,iyba->xy', t2a.conj(), t2b)
    return -dm1occ, dm1vir


class KsymAdaptedKMP2(kmp2.KMP2):
    def kernel(self, mo_energy=None, mo_coeff=None, with_t2=WITH_T2):
        if mo_energy is None: mo_energy = self.mo_energy
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if mo_energy is None or mo_coeff is None:
            logger.warn('mo_coeff, mo_energy are not given.\n'
                        'You may need to call mf.kernel() to generate them.')
            raise RuntimeError

        mo_coeff, mo_energy = kmp2._add_padding(self, mo_coeff, mo_energy)

        # TODO: compute e_hf for non-canonical SCF
        self.e_hf = self._scf.e_tot

        self.e_corr, self.t2 = \
                kernel(self, mo_energy, mo_coeff, verbose=self.verbose, with_t2=with_t2)

        self.e_corr_ss = getattr(self.e_corr, 'e_corr_ss', 0)
        self.e_corr_os = getattr(self.e_corr, 'e_corr_os', 0)
        self.e_corr = float(self.e_corr)

        self._finalize()

        return self.e_corr, self.t2

    make_rdm1 = make_rdm1

    def make_rdm2(self):
        raise NotImplementedError

KRMP2 = KMP2 = KsymAdaptedKMP2

from pyscf.pbc import scf
scf.khf_ksymm.KRHF.MP2 = lib.class_as_method(KRMP2)

if __name__ == '__main__':
    from pyscf.pbc import gto, scf, mp

    cell = gto.Cell()
    cell.atom='''
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    '''
    cell.basis = 'gth-szv'
    cell.pseudo = 'gth-pade'
    cell.a = '''
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000'''
    cell.unit = 'B'
    cell.verbose = 5
    cell.build()

    kpts = cell.make_kpts([2,2,2], space_group_symmetry=True)
    kmf = scf.KRHF(cell, kpts=kpts, exxdiv=None)
    ehf = kmf.kernel()

    mymp = mp.KMP2(kmf)
    emp2, t2 = mymp.kernel()
    print(emp2 - -0.13314158977189)



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyscf.solvent import ddcosmo
from pyscf.solvent import pcm
from pyscf.solvent import smd

def ddCOSMO(method_or_mol, solvent_obj=None, dm=None):
    '''Initialize ddCOSMO model.

    Examples:

    >>> mf = ddCOSMO(scf.RHF(mol))
    >>> mf.kernel()
    >>> sol = ddCOSMO(mol)
    >>> mc = ddCOSMO(CASCI(mf, 6, 6), sol)
    >>> mc.kernel()
    '''
    from pyscf import gto
    from pyscf import scf, mcscf
    from pyscf import tdscf

    if isinstance(method_or_mol, gto.mole.Mole):
        return ddcosmo.DDCOSMO(method_or_mol)

    method = method_or_mol
    if isinstance(method, scf.hf.SCF):
        return ddcosmo.ddcosmo_for_scf(method, solvent_obj, dm)
    elif isinstance(method, mcscf.casci.CASBase):
        if isinstance(method, mcscf.mc1step.CASSCF):
            return ddcosmo.ddcosmo_for_casscf(method, solvent_obj, dm)
        elif isinstance(method, mcscf.casci.CASCI):
            return ddcosmo.ddcosmo_for_casci(method, solvent_obj, dm)
    elif isinstance(method, tdscf.rhf.TDBase):
        return ddcosmo.ddcosmo_for_tdscf(method, solvent_obj, dm)
    elif hasattr(method, '_scf'):
        return ddcosmo.ddcosmo_for_post_scf(method, solvent_obj, dm)
    raise RuntimeError(f'ddCOSMO for {method} not available')
DDCOSMO = ddCOSMO

def ddPCM(method_or_mol, solvent_obj=None, dm=None):
    '''Initialize ddPCM model.

    Examples:

    >>> mf = ddPCM(scf.RHF(mol))
    >>> mf.kernel()
    >>> sol = ddPCM(mol)
    >>> mc = ddPCM(CASCI(mf, 6, 6), sol)
    >>> mc.kernel()
    '''
    from pyscf import gto
    from pyscf import scf, mcscf
    from pyscf import tdscf
    from pyscf.solvent import ddpcm

    if isinstance(method_or_mol, gto.mole.Mole):
        return ddpcm.DDPCM(method_or_mol)

    method = method_or_mol
    if isinstance(method, scf.hf.SCF):
        return ddpcm.ddpcm_for_scf(method, solvent_obj, dm)
    elif isinstance(method, mcscf.casci.CASBase):
        if isinstance(method, mcscf.mc1step.CASSCF):
            return ddpcm.ddpcm_for_casscf(method, solvent_obj, dm)
        elif isinstance(method, mcscf.casci.CASCI):
            return ddpcm.ddpcm_for_casci(method, solvent_obj, dm)
    elif isinstance(method, tdscf.rhf.TDBase):
        return ddpcm.ddpcm_for_tdscf(method, solvent_obj, dm)
    elif hasattr(method, '_scf'):
        return ddpcm.ddpcm_for_post_scf(method, solvent_obj, dm)
    raise RuntimeError(f'ddPCM for {method} not available')
DDPCM = ddPCM

def PE(method_or_mol, solvent_obj=None, dm=None):
    '''Initialize polarizable embedding model.

    Args:
        method_or_mol (pyscf method object or gto.Mole object)
            If method_or_mol is gto.Mole object, this function returns a
            PolEmbed object constructed with this Mole object.
        solvent_obj (PolEmbed object or dictionary with options or str)
            If solvent_obj is an object of PolEmbed class, the PE-enabled
            method will be created using solvent_obj.
            If solvent_obj is dict or str, a PolEmbed object will
            be created first with the solvent_obj, on top of which PE-enabled
            method will be created.

    Examples:

    >>> pe_options = {"potfile": "pyframe.pot"}
    >>> mf = PE(scf.RHF(mol), pe_options)
    >>> mf.kernel()
    '''
    from pyscf.solvent import pol_embed
    from pyscf import gto, scf, mcscf, tdscf

    if isinstance(method_or_mol, gto.mole.Mole):
        return pol_embed.PolEmbed(method_or_mol, solvent_obj)

    method = method_or_mol
    if isinstance(method, scf.hf.SCF):
        assert solvent_obj is not None
        return pol_embed.pe_for_scf(method, solvent_obj, dm)

    if solvent_obj is None:
        solvent_obj = method._scf.with_solvent
    if isinstance(method, mcscf.casci.CASBase):
        if isinstance(method, mcscf.mc1step.CASSCF):
            return pol_embed.pe_for_casscf(method, solvent_obj, dm)
        elif isinstance(method, mcscf.casci.CASCI):
            return pol_embed.pe_for_casci(method, solvent_obj, dm)
    elif isinstance(method, tdscf.rhf.TDBase):
        return pol_embed.pe_for_tdscf(method, solvent_obj, dm)
    elif hasattr(method, '_scf'):
        return pol_embed.pe_for_post_scf(method, solvent_obj, dm)
    raise RuntimeError(f'PolEmbed for {method} not available')

def PCM(method_or_mol, solvent_obj=None, dm=None):
    '''Initialize PCM model.

    Examples:

    >>> mf = PCM(scf.RHF(mol))
    >>> mf.kernel()
    >>> sol = PCM(mol)
    >>> mc = PCM(CASCI(mf, 6, 6), sol)
    >>> mc.kernel()
    '''
    from pyscf import gto
    from pyscf import scf, mcscf
    from pyscf import tdscf

    if isinstance(method_or_mol, gto.mole.Mole):
        return pcm.PCM(method_or_mol)

    method = method_or_mol
    if isinstance(method, scf.hf.SCF):
        return pcm.pcm_for_scf(method, solvent_obj, dm)
    elif isinstance(method, mcscf.casci.CASBase):
        if isinstance(method, mcscf.mc1step.CASSCF):
            return pcm.pcm_for_casscf(method, solvent_obj, dm)
        elif isinstance(method, mcscf.casci.CASCI):
            return pcm.pcm_for_casci(method, solvent_obj, dm)
    elif isinstance(method, tdscf.rhf.TDBase):
        return pcm.pcm_for_tdscf(method, solvent_obj, dm)
    elif hasattr(method, '_scf'):
        return pcm.pcm_for_post_scf(method, solvent_obj, dm)
    raise RuntimeError(f'PCM for {method} not available')

PCM = PCM

def SMD(method_or_mol, solvent_obj=None, dm=None):
    '''Initialize PCM model.

    Examples:

    >>> mf = PCM(scf.RHF(mol))
    >>> mf.kernel()
    >>> sol = PCM(mol)
    >>> mc = PCM(CASCI(mf, 6, 6), sol)
    >>> mc.kernel()
    '''
    from pyscf import gto
    from pyscf import scf

    method = method_or_mol
    if isinstance(method, gto.mole.Mole):
        return smd.SMD(method)
    elif isinstance(method, scf.hf.SCF):
        return smd.smd_for_scf(method, solvent_obj, dm)
    raise RuntimeError(f'SMD for {method} not available')

SMD = SMD



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver Backhouse <olbackhouse@gmail.com>
#         George Booth <george.booth@kcl.ac.uk>
#

'''
MPI helper functions using mpi4py
'''

import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__

INT_MAX = 2147483647
BLKSIZE = INT_MAX // 32 + 1

# attempt to successfully load and init the MPI, else assume 1 core:
try:
    from mpi4py import MPI as mpi
    comm = mpi.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()
except Exception:
    mpi = None
    comm = None
    size = 1
    rank = 0

SCALE_PRANGE_STEP = False


def bcast(buf, root=0):
    if size == 1:
        return buf

    is_array = isinstance(buf, np.ndarray)
    buf = np.asarray(buf, order='C')
    buf = buf.astype(buf.dtype.char)
    shape, mpi_dtype = comm.bcast((buf.shape, buf.dtype.char))

    if rank != root:
        buf = np.empty(shape, dtype=mpi_dtype)

    buf_seg = np.ndarray(buf.size, dtype=buf.dtype, buffer=buf)
    for p0, p1 in lib.prange(0, buf.size, BLKSIZE):
        comm.Bcast(buf_seg[p0:p1], root)

    return buf if is_array else buf.ravel()[0]


def bcast_dict(buf, root=0):
    if size == 1:
        return buf

    buf = comm.bcast(buf, root)

    return buf


def reduce(sendbuf, root=0, op=getattr(mpi, 'SUM', None)):
    if size == 1:
        return sendbuf

    is_array = isinstance(sendbuf, np.ndarray)
    sendbuf = np.asarray(sendbuf, order='C')
    sendbuf = sendbuf.astype(sendbuf.dtype.char)
    shape, mpi_dtype = comm.bcast((sendbuf.shape, sendbuf.dtype.char))
    assert sendbuf.shape == shape and sendbuf.dtype.char == mpi_dtype

    recvbuf = np.zeros_like(sendbuf)
    send_seg = np.ndarray(sendbuf.size, dtype=sendbuf.dtype, buffer=sendbuf)
    recv_seg = np.ndarray(recvbuf.size, dtype=recvbuf.dtype, buffer=recvbuf)
    for p0, p1 in lib.prange(0, sendbuf.size, BLKSIZE):
        comm.Reduce(send_seg[p0:p1], recv_seg[p0:p1], op, root)

    if rank == root:
        return recvbuf if is_array else recvbuf.ravel()[0]
    else:
        return sendbuf if is_array else sendbuf.ravel()[0]


def allreduce(sendbuf, root=0, op=getattr(mpi, 'SUM', None)):
    if size == 1:
        return sendbuf

    is_array = isinstance(sendbuf, np.ndarray)
    sendbuf = np.asarray(sendbuf, order='C')
    sendbuf = sendbuf.astype(sendbuf.dtype.char)
    shape, mpi_dtype = comm.bcast((sendbuf.shape, sendbuf.dtype.char))
    assert sendbuf.shape == shape and sendbuf.dtype.char == mpi_dtype

    recvbuf = np.zeros_like(sendbuf)
    send_seg = np.ndarray(sendbuf.size, dtype=sendbuf.dtype, buffer=sendbuf)
    recv_seg = np.ndarray(recvbuf.size, dtype=recvbuf.dtype, buffer=recvbuf)
    for p0, p1 in lib.prange(0, sendbuf.size, BLKSIZE):
        comm.Allreduce(send_seg[p0:p1], recv_seg[p0:p1], op)

    return recvbuf if is_array else recvbuf.ravel()[0]


def allreduce_safe_inplace(array):
    if size == 1:
        return array

    from pyscf.pbc.mpitools.mpi_helper import safeAllreduceInPlace

    safeAllreduceInPlace(comm, array)


def barrier():
    if comm is not None:
        comm.Barrier()


def nrange(start, stop=None, step=1):
    if stop is None:
        start, stop = 0, start

    yield from range(start+rank, stop, step*size)


def prange(start, stop, step):
    ''' :func:`lib.prange` distributed over MPI processes. Returns
        the range for a single MPI rank.
    '''

    if size == 1:
        for p0, p1 in lib.prange(start, stop, step):
            yield p0, p1
    else:
        if SCALE_PRANGE_STEP:
            step //= size

        split = lambda x : x * (stop-start) // size

        start0 = split(rank)
        stop0 = stop if rank == (size-1) else split(rank+1)

        for p0, p1 in lib.prange(start0, stop0, step):
            yield p0, p1



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Ref:
# J. Chem. Phys. 117, 7433
#


from functools import reduce
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.dft import rks
from pyscf.dft import numint
from pyscf.scf import cphf
from pyscf.grad import rks as rks_grad
from pyscf.grad import tdrhf


#
# Given Y = 0, TDDFT gradients (XAX+XBY+YBX+YAY)^1 turn to TDA gradients (XAX)^1
#
def grad_elec(td_grad, x_y, singlet=True, atmlst=None,
              max_memory=2000, verbose=logger.INFO):
    '''
    Electronic part of TDA, TDDFT nuclear gradients

    Args:
        td_grad : grad.tdrhf.Gradients or grad.tdrks.Gradients object.

        x_y : a two-element list of numpy arrays
            TDDFT X and Y amplitudes. If Y is set to 0, this function computes
            TDA energy gradients.
    '''
    log = logger.new_logger(td_grad, verbose)
    time0 = logger.process_clock(), logger.perf_counter()

    assert td_grad.base.frozen is None

    mol = td_grad.mol
    mf = td_grad.base._scf
    mo_coeff = mf.mo_coeff
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    nao, nmo = mo_coeff.shape
    nocc = (mo_occ>0).sum()
    nvir = nmo - nocc
    x, y = x_y
    xpy = (x+y).reshape(nocc,nvir).T
    xmy = (x-y).reshape(nocc,nvir).T
    orbv = mo_coeff[:,nocc:]
    orbo = mo_coeff[:,:nocc]

    dvv = numpy.einsum('ai,bi->ab', xpy, xpy) + numpy.einsum('ai,bi->ab', xmy, xmy)
    doo =-numpy.einsum('ai,aj->ij', xpy, xpy) - numpy.einsum('ai,aj->ij', xmy, xmy)
    dmxpy = reduce(numpy.dot, (orbv, xpy, orbo.T))
    dmxmy = reduce(numpy.dot, (orbv, xmy, orbo.T))
    dmzoo = reduce(numpy.dot, (orbo, doo, orbo.T))
    dmzoo+= reduce(numpy.dot, (orbv, dvv, orbv.T))

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, td_grad.max_memory*.9-mem_now)

    ni = mf._numint
    ni.libxc.test_deriv_order(mf.xc, 3, raise_error=True)
    omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, mol.spin)
    f1vo, f1oo, vxc1, k1ao = \
            _contract_xc_kernel(td_grad, mf.xc, dmxpy,
                                dmzoo, True, True, singlet, max_memory)

    if ni.libxc.is_hybrid_xc(mf.xc):
        dm = (dmzoo, dmxpy+dmxpy.T, dmxmy-dmxmy.T)
        vj, vk = mf.get_jk(mol, dm, hermi=0)
        vk *= hyb
        if omega != 0:
            vk += mf.get_k(mol, dm, hermi=0, omega=omega) * (alpha-hyb)
        veff0doo = vj[0] * 2 - vk[0] + f1oo[0] + k1ao[0] * 2
        wvo = reduce(numpy.dot, (orbv.T, veff0doo, orbo)) * 2
        if singlet:
            veff = vj[1] * 2 - vk[1] + f1vo[0] * 2
        else:
            veff = f1vo[0] - vk[1]
        veff0mop = reduce(numpy.dot, (mo_coeff.T, veff, mo_coeff))
        wvo -= numpy.einsum('ki,ai->ak', veff0mop[:nocc,:nocc], xpy) * 2
        wvo += numpy.einsum('ac,ai->ci', veff0mop[nocc:,nocc:], xpy) * 2
        veff = -vk[2]
        veff0mom = reduce(numpy.dot, (mo_coeff.T, veff, mo_coeff))
        wvo -= numpy.einsum('ki,ai->ak', veff0mom[:nocc,:nocc], xmy) * 2
        wvo += numpy.einsum('ac,ai->ci', veff0mom[nocc:,nocc:], xmy) * 2
    else:
        vj = mf.get_j(mol, (dmzoo, dmxpy+dmxpy.T), hermi=1)
        veff0doo = vj[0] * 2 + f1oo[0] + k1ao[0] * 2
        wvo = reduce(numpy.dot, (orbv.T, veff0doo, orbo)) * 2
        if singlet:
            veff = vj[1] * 2 + f1vo[0] * 2
        else:
            veff = f1vo[0]
        veff0mop = reduce(numpy.dot, (mo_coeff.T, veff, mo_coeff))
        wvo -= numpy.einsum('ki,ai->ak', veff0mop[:nocc,:nocc], xpy) * 2
        wvo += numpy.einsum('ac,ai->ci', veff0mop[nocc:,nocc:], xpy) * 2
        veff0mom = numpy.zeros((nmo,nmo))

    # set singlet=None, generate function for CPHF type response kernel
    vresp = mf.gen_response(singlet=None, hermi=1,
                            with_nlc=not td_grad.base.exclude_nlc)
    def fvind(x):
        dm = reduce(numpy.dot, (orbv, x.reshape(nvir,nocc)*2, orbo.T))
        v1ao = vresp(dm+dm.T)
        return reduce(numpy.dot, (orbv.T, v1ao, orbo)).ravel()
    z1 = cphf.solve(fvind, mo_energy, mo_occ, wvo,
                    max_cycle=td_grad.cphf_max_cycle,
                    tol=td_grad.cphf_conv_tol)[0]
    z1 = z1.reshape(nvir,nocc)
    time1 = log.timer('Z-vector using CPHF solver', *time0)

    z1ao = reduce(numpy.dot, (orbv, z1, orbo.T))
    veff = vresp(z1ao+z1ao.T)

    im0 = numpy.zeros((nmo,nmo))
    im0[:nocc,:nocc] = reduce(numpy.dot, (orbo.T, veff0doo+veff, orbo))
    im0[:nocc,:nocc]+= numpy.einsum('ak,ai->ki', veff0mop[nocc:,:nocc], xpy)
    im0[:nocc,:nocc]+= numpy.einsum('ak,ai->ki', veff0mom[nocc:,:nocc], xmy)
    im0[nocc:,nocc:] = numpy.einsum('ci,ai->ac', veff0mop[nocc:,:nocc], xpy)
    im0[nocc:,nocc:]+= numpy.einsum('ci,ai->ac', veff0mom[nocc:,:nocc], xmy)
    im0[nocc:,:nocc] = numpy.einsum('ki,ai->ak', veff0mop[:nocc,:nocc], xpy)*2
    im0[nocc:,:nocc]+= numpy.einsum('ki,ai->ak', veff0mom[:nocc,:nocc], xmy)*2

    zeta = lib.direct_sum('i+j->ij', mo_energy, mo_energy) * .5
    zeta[nocc:,:nocc] = mo_energy[:nocc]
    zeta[:nocc,nocc:] = mo_energy[nocc:]
    dm1 = numpy.zeros((nmo,nmo))
    dm1[:nocc,:nocc] = doo
    dm1[nocc:,nocc:] = dvv
    dm1[nocc:,:nocc] = z1
    dm1[:nocc,:nocc] += numpy.eye(nocc)*2 # for ground state
    im0 = reduce(numpy.dot, (mo_coeff, im0+zeta*dm1, mo_coeff.T))

    # Initialize hcore_deriv with the underlying SCF object because some
    # extensions (e.g. QM/MM, solvent) modifies the SCF object only.
    mf_grad = td_grad.base._scf.nuc_grad_method()
    hcore_deriv = mf_grad.hcore_generator(mol)
    s1 = mf_grad.get_ovlp(mol)

    dmz1doo = z1ao + dmzoo
    oo0 = reduce(numpy.dot, (orbo, orbo.T))
    if ni.libxc.is_hybrid_xc(mf.xc):
        dm = (oo0, dmz1doo+dmz1doo.T, dmxpy+dmxpy.T, dmxmy-dmxmy.T)
        vj, vk = td_grad.get_jk(mol, dm)
        vk *= hyb
        if omega != 0:
            vk += td_grad.get_k(mol, dm, omega=omega) * (alpha-hyb)
        vj = vj.reshape(-1,3,nao,nao)
        vk = vk.reshape(-1,3,nao,nao)
        veff1 = -vk
        if singlet:
            veff1 += vj * 2
        else:
            veff1[:2] += vj[:2] * 2
    else:
        vj = td_grad.get_j(mol, (oo0, dmz1doo+dmz1doo.T, dmxpy+dmxpy.T))
        vj = vj.reshape(-1,3,nao,nao)
        veff1 = numpy.zeros((4,3,nao,nao))
        if singlet:
            veff1[:3] = vj * 2
        else:
            veff1[:2] = vj[:2] * 2

    fxcz1 = _contract_xc_kernel(td_grad, mf.xc, z1ao, None,
                                False, False, True, max_memory)[0]

    veff1[0] += vxc1[1:]
    veff1[1] +=(f1oo[1:] + fxcz1[1:] + k1ao[1:]*2)*2 # *2 for dmz1doo+dmz1oo.T
    if singlet:
        veff1[2] += f1vo[1:] * 2
    else:
        veff1[2] += f1vo[1:]
    time1 = log.timer('2e AO integral derivatives', *time1)

    if atmlst is None:
        atmlst = range(mol.natm)
    offsetdic = mol.offset_nr_by_atom()
    de = numpy.zeros((len(atmlst),3))
    for k, ia in enumerate(atmlst):
        shl0, shl1, p0, p1 = offsetdic[ia]

        # Ground state gradients
        h1ao = hcore_deriv(ia)
        h1ao[:,p0:p1]   += veff1[0,:,p0:p1]
        h1ao[:,:,p0:p1] += veff1[0,:,p0:p1].transpose(0,2,1)
        # oo0*2 for doubly occupied orbitals
        e1  = numpy.einsum('xpq,pq->x', h1ao, oo0) * 2

        e1 += numpy.einsum('xpq,pq->x', h1ao, dmz1doo)
        e1 -= numpy.einsum('xpq,pq->x', s1[:,p0:p1], im0[p0:p1])
        e1 -= numpy.einsum('xqp,pq->x', s1[:,p0:p1], im0[:,p0:p1])

        e1 += numpy.einsum('xij,ij->x', veff1[1,:,p0:p1], oo0[p0:p1])
        e1 += numpy.einsum('xij,ij->x', veff1[2,:,p0:p1], dmxpy[p0:p1,:]) * 2
        e1 += numpy.einsum('xji,ij->x', veff1[2,:,p0:p1], dmxpy[:,p0:p1]) * 2
        e1 += numpy.einsum('xij,ij->x', veff1[3,:,p0:p1], dmxmy[p0:p1,:]) * 2
        e1 -= numpy.einsum('xji,ij->x', veff1[3,:,p0:p1], dmxmy[:,p0:p1]) * 2

        e1 += td_grad.extra_force(ia, locals())

        de[k] = e1

    log.timer('TDRKS nuclear gradients', *time0)
    return de


# dmvo, dmoo in AO-representation
# Note spin-trace is applied for fxc, kxc
#TODO: to include the response of grids
def _contract_xc_kernel(td_grad, xc_code, dmvo, dmoo=None, with_vxc=True,
                        with_kxc=True, singlet=True, max_memory=2000):
    mol = td_grad.mol
    mf = td_grad.base._scf
    grids = mf.grids

    ni = mf._numint
    xctype = ni._xc_type(xc_code)

    mo_coeff = mf.mo_coeff
    mo_occ = mf.mo_occ
    nao, nmo = mo_coeff.shape
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()

    # dmvo ~ reduce(numpy.dot, (orbv, Xai, orbo.T))
    dmvo = (dmvo + dmvo.T) * .5 # because K_{ia,jb} == K_{ia,bj}

    f1vo = numpy.zeros((4,nao,nao))  # 0th-order, d/dx, d/dy, d/dz
    deriv = 2
    if dmoo is not None:
        f1oo = numpy.zeros((4,nao,nao))
    else:
        f1oo = None
    if with_vxc:
        v1ao = numpy.zeros((4,nao,nao))
    else:
        v1ao = None
    if with_kxc:
        k1ao = numpy.zeros((4,nao,nao))
        deriv = 3
    else:
        k1ao = None

    if xctype == 'HF':
        return f1vo, f1oo, v1ao, k1ao
    elif xctype == 'LDA':
        fmat_, ao_deriv = _lda_eval_mat_, 1
    elif xctype == 'GGA':
        fmat_, ao_deriv = _gga_eval_mat_, 2
    elif xctype == 'MGGA':
        fmat_, ao_deriv = _mgga_eval_mat_, 2
        logger.warn(td_grad, 'TDRKS-MGGA Gradients may be inaccurate due to grids response')
    else:
        raise NotImplementedError(f'td-rks for functional {xc_code}')

    if mf.do_nlc():
        raise NotImplementedError("TDDFT gradient with NLC contribution is not supported yet. "
                                  "Please set exclude_nlc field of tdscf object to True, "
                                  "which will turn off NLC contribution in the whole TDDFT calculation.")

    if singlet:
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            if xctype == 'LDA':
                ao0 = ao[0]
            else:
                ao0 = ao
            rho = ni.eval_rho2(mol, ao0, mo_coeff, mo_occ, mask, xctype, with_lapl=False)
            vxc, fxc, kxc = ni.eval_xc_eff(xc_code, rho, deriv, xctype=xctype)[1:]

            rho1 = ni.eval_rho(mol, ao0, dmvo, mask, xctype, hermi=1,
                               with_lapl=False) * 2  # *2 for alpha + beta
            if xctype == 'LDA':
                rho1 = rho1[numpy.newaxis]
            wv = numpy.einsum('yg,xyg,g->xg', rho1, fxc, weight)
            fmat_(mol, f1vo, ao, wv, mask, shls_slice, ao_loc)

            if dmoo is not None:
                rho2 = ni.eval_rho(mol, ao0, dmoo, mask, xctype, hermi=1, with_lapl=False) * 2
                if xctype == 'LDA':
                    rho2 = rho2[numpy.newaxis]
                wv = numpy.einsum('yg,xyg,g->xg', rho2, fxc, weight)
                fmat_(mol, f1oo, ao, wv, mask, shls_slice, ao_loc)
            if with_vxc:
                fmat_(mol, v1ao, ao, vxc * weight, mask, shls_slice, ao_loc)
            if with_kxc:
                wv = numpy.einsum('yg,zg,xyzg,g->xg', rho1, rho1, kxc, weight)
                fmat_(mol, k1ao, ao, wv, mask, shls_slice, ao_loc)
    else:
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            if xctype == 'LDA':
                ao0 = ao[0]
            else:
                ao0 = ao
            rho = ni.eval_rho2(mol, ao0, mo_coeff, mo_occ, mask, xctype, with_lapl=False)
            rho *= .5
            rho = numpy.repeat(rho[numpy.newaxis], 2, axis=0)
            vxc, fxc, kxc = ni.eval_xc_eff(xc_code, rho, deriv, xctype=xctype)[1:]
            # fxc_t couples triplet excitation amplitudes
            # 1/2 int (tia - tIA) fxc (tjb - tJB) = tia fxc_t tjb
            fxc_t = fxc[:,:,0] - fxc[:,:,1]
            fxc_t = fxc_t[0] - fxc_t[1]

            rho1 = ni.eval_rho(mol, ao0, dmvo, mask, xctype, hermi=1, with_lapl=False)
            if xctype == 'LDA':
                rho1 = rho1[numpy.newaxis]
            wv = numpy.einsum('yg,xyg,g->xg', rho1, fxc_t, weight)
            fmat_(mol, f1vo, ao, wv, mask, shls_slice, ao_loc)

            if dmoo is not None:
                # fxc_s == 2 * fxc of spin restricted xc kernel
                # provides f1oo to couple the interaction between first order MO
                # and density response of tddft amplitudes, which is described by dmoo
                fxc_s = fxc[0,:,0] + fxc[0,:,1]
                rho2 = ni.eval_rho(mol, ao0, dmoo, mask, xctype, hermi=1, with_lapl=False)
                if xctype == 'LDA':
                    rho2 = rho2[numpy.newaxis]
                wv = numpy.einsum('yg,xyg,g->xg', rho2, fxc_s, weight)
                fmat_(mol, f1oo, ao, wv, mask, shls_slice, ao_loc)
            if with_vxc:
                vxc = vxc[0]
                fmat_(mol, v1ao, ao, vxc * weight, mask, shls_slice, ao_loc)
            if with_kxc:
                # kxc in terms of the triplet coupling
                # 1/2 int (tia - tIA) kxc (tjb - tJB) = tia kxc_t tjb
                kxc = kxc[0,:,0] - kxc[0,:,1]
                kxc = kxc[:,:,0] - kxc[:,:,1]
                wv = numpy.einsum('yg,zg,xyzg,g->xg', rho1, rho1, kxc, weight)
                fmat_(mol, k1ao, ao, wv, mask, shls_slice, ao_loc)

    f1vo[1:] *= -1
    if f1oo is not None: f1oo[1:] *= -1
    if v1ao is not None: v1ao[1:] *= -1
    if k1ao is not None: k1ao[1:] *= -1
    return f1vo, f1oo, v1ao, k1ao

def _lda_eval_mat_(mol, vmat, ao, wv, mask, shls_slice, ao_loc):
    aow = numint._scale_ao(ao[0], wv[0])
    for k in range(4):
        vmat[k] += numint._dot_ao_ao(mol, ao[k], aow, mask, shls_slice, ao_loc)
    return vmat

def _gga_eval_mat_(mol, vmat, ao, wv, mask, shls_slice, ao_loc):
    wv[0] *= .5  # *.5 because vmat + vmat.T at the end
    aow = numint._scale_ao(ao[:4], wv[:4])
    tmp = numint._dot_ao_ao(mol, ao[0], aow, mask, shls_slice, ao_loc)
    vmat[0] += tmp + tmp.T
    rks_grad._gga_grad_sum_(vmat[1:], mol, ao, wv, mask, ao_loc)
    return vmat

def _mgga_eval_mat_(mol, vmat, ao, wv, mask, shls_slice, ao_loc):
    wv[0] *= .5  # *.5 because vmat + vmat.T at the end
    wv[4] *= .5  # *.5 for 1/2 in tau
    aow = numint._scale_ao(ao[:4], wv[:4])
    tmp = numint._dot_ao_ao(mol, ao[0], aow, mask, shls_slice, ao_loc)
    vmat[0] += tmp + tmp.T
    vmat[0] += numint._tau_dot(mol, ao, ao, wv[4], mask, shls_slice, ao_loc)
    rks_grad._gga_grad_sum_(vmat[1:], mol, ao, wv[:4], mask, ao_loc)
    rks_grad._tau_grad_dot_(vmat[1:], mol, ao, wv[4], mask, ao_loc, True)
    return vmat


class Gradients(tdrhf.Gradients):
    @lib.with_doc(grad_elec.__doc__)
    def grad_elec(self, xy, singlet, atmlst=None):
        return grad_elec(self, xy, singlet, atmlst, self.max_memory, self.verbose)

Grad = Gradients

from pyscf import tdscf
tdscf.rks.TDA.Gradients = tdscf.rks.TDDFT.Gradients = lib.class_as_method(Gradients)



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyscf.pbc.cc import ccsd
from pyscf.pbc.cc import kccsd_rhf as krccsd
from pyscf.pbc.cc import kccsd_uhf as kuccsd
from pyscf.pbc.cc import kccsd     as kgccsd
from pyscf.pbc.cc import eom_kccsd_rhf
from pyscf.pbc.cc import eom_kccsd_uhf
from pyscf.pbc.cc import eom_kccsd_ghf
from pyscf.pbc.cc.kccsd_rhf_ksymm import KsymAdaptedRCCSD

def RCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    mf = mf.to_rhf()
    return ccsd.RCCSD(mf, frozen, mo_coeff, mo_occ)

CCSD = RCCSD

def UCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    mf = mf.to_uhf()
    return ccsd.UCCSD(mf, frozen, mo_coeff, mo_occ)

def GCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    mf = mf.to_ghf()
    return ccsd.GCCSD(mf, frozen, mo_coeff, mo_occ)

def KGCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    mf = mf.to_ghf()
    return kgccsd.GCCSD(mf, frozen, mo_coeff, mo_occ)

def KRCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.pbc import scf
    assert isinstance(mf, scf.khf.KSCF)
    if not isinstance(mf, scf.khf.KRHF):
        mf = mf.to_rhf()
    return krccsd.RCCSD(mf, frozen, mo_coeff, mo_occ)

KCCSD = KRCCSD

def KUCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.pbc import scf
    assert isinstance(mf, scf.khf.KSCF)
    if not isinstance(mf, scf.kuhf.KUHF):
        mf = mf.to_uhf()
    return kuccsd.UCCSD(mf, frozen, mo_coeff, mo_occ)



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import ctypes
import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad import rhf as mol_rhf
from pyscf.grad.rhf import _write
from pyscf.pbc.gto.pseudo import pp_int
from pyscf.pbc.lib.kpts_helper import gamma_point
from pyscf.pbc.dft.multigrid import MultiGridNumInt2

SCREEN_VHF_DM_CONTRA = getattr(__config__, 'pbc_rhf_grad_screen_vhf_dm_contract', True)
libpbc = lib.load_library('libpbc')

def grad_elec(mf_grad, mo_energy=None, mo_coeff=None, mo_occ=None,
              atmlst=None, kpt=np.zeros(3)):
    mf = mf_grad.base
    mol = mf_grad.mol
    if mo_energy is None: mo_energy = mf.mo_energy
    if mo_occ is None:    mo_occ = mf.mo_occ
    if mo_coeff is None:  mo_coeff = mf.mo_coeff
    log = logger.Logger(mf_grad.stdout, mf_grad.verbose)

    if hasattr(mf, '_numint'):
        ni = mf._numint
        assert isinstance(ni, MultiGridNumInt2)
    else:
        ni = mf.with_df
        raise NotImplementedError

    s1 = mf_grad.get_ovlp(mol, kpt)
    dm0 = mf.make_rdm1(mo_coeff, mo_occ)

    t0 = (logger.process_clock(), logger.perf_counter())
    log.debug('Computing Gradients of NR-HF Coulomb repulsion')
    vhf = mf_grad.get_veff(mol, dm0, kpt)
    log.timer('gradients of 2e part', *t0)

    dme0 = mf_grad.make_rdm1e(mo_energy, mo_coeff, mo_occ)

    if atmlst is None:
        atmlst = range(mol.natm)

    if gamma_point(kpt):
        h1ao = -mol.pbc_intor('int1e_ipkin', kpt=kpt)
        if mol._pseudo:
            de  = ni.vpploc_part1_nuc_grad(dm0, kpts=kpt.reshape(-1,3))
            de += pp_int.vpploc_part2_nuc_grad(mol, dm0)
            de += pp_int.vppnl_nuc_grad(mol, dm0)
            if hasattr(ni, 'vpplocG_part1'):
                if ni.vpplocG_part1 is None:
                    h1ao -= ni.get_vpploc_part1_ip1(kpts=kpt.reshape(-1,3))
        else:
            de = ni.get_nuc_nuc_grad(dm0, kpts=kpt)
            h1ao -= ni.get_nuc_ip1(kpts=kpt)
        de += _contract_vhf_dm(mf_grad, np.add(h1ao, vhf), dm0) * 2
        de += _contract_vhf_dm(mf_grad, s1, dme0) * -2
        h1ao = s1 = vhf = dm0 = dme0 = None
        de = de[atmlst]
    else:
        raise NotImplementedError

    for k, ia in enumerate(atmlst):
        de[k] += mf_grad.extra_force(ia, locals())

    if log.verbose >= logger.DEBUG:
        log.debug('gradients of electronic part')
        _write(log, mol, de, atmlst)
    return de


def _contract_vhf_dm(mf_grad, vhf, dm, comp=3, atmlst=None,
                     screen=SCREEN_VHF_DM_CONTRA):
    from pyscf.gto.mole import ao_loc_nr, ATOM_OF
    from pyscf.pbc.gto import build_neighbor_list_for_shlpairs, free_neighbor_list

    t0 = (logger.process_clock(), logger.perf_counter())

    mol = mf_grad.mol
    natm = mol.natm
    nbas = mol.nbas
    shls_slice = np.asarray([0,nbas,0,nbas], order="C", dtype=np.int32)
    ao_loc = np.asarray(ao_loc_nr(mol), order="C", dtype=np.int32)
    shls_atm = np.asarray(mol._bas[:,ATOM_OF].copy(), order="C", dtype=np.int32)

    de = np.zeros((natm,comp), order="C")
    vhf = np.asarray(vhf, order="C")
    dm = np.asarray(dm, order="C")

    if screen:
        neighbor_list = build_neighbor_list_for_shlpairs(mol)
    else:
        neighbor_list = lib.c_null_ptr()
    func = getattr(libpbc, "contract_vhf_dm", None)
    try:
        func(de.ctypes.data_as(ctypes.c_void_p),
             vhf.ctypes.data_as(ctypes.c_void_p),
             dm.ctypes.data_as(ctypes.c_void_p),
             ctypes.byref(neighbor_list),
             shls_slice.ctypes.data_as(ctypes.c_void_p),
             ao_loc.ctypes.data_as(ctypes.c_void_p),
             shls_atm.ctypes.data_as(ctypes.c_void_p),
             ctypes.c_int(comp), ctypes.c_int(natm),
             ctypes.c_int(nbas))
    except RuntimeError:
        raise
    free_neighbor_list(neighbor_list)

    if atmlst is not None:
        de = de[atmlst]

    logger.timer(mf_grad, '_contract_vhf_dm', *t0)
    return de


def get_ovlp(cell, kpt=np.zeros(3)):
    return -cell.pbc_intor('int1e_ipovlp', kpt=kpt)


def get_veff(mf_grad, mol, dm, kpt=np.zeros(3)):
    mf = mf_grad.base
    xc_code = getattr(mf, 'xc', None)
    kpts = kpt.reshape(-1,3)
    return -mf._numint.get_veff_ip1(dm, xc_code=xc_code, kpts=kpts)


def grad_nuc(cell, atmlst=None, ew_eta=None, ew_cut=None):
    from pyscf.pbc.gto import ewald_methods

    t0 = (logger.process_clock(), logger.perf_counter())

    grad = ewald_methods.ewald_nuc_grad(cell, ew_eta, ew_cut)
    if atmlst is not None:
        grad = grad[atmlst]

    logger.timer(cell, 'nuclear gradient', *t0)
    return grad


class GradientsBase(mol_rhf.GradientsBase):
    '''Base class for Gamma-point nuclear gradient'''
    def grad_nuc(self, mol=None, atmlst=None):
        if mol is None: mol = self.mol
        return grad_nuc(mol, atmlst)

    def get_ovlp(self, mol=None, kpt=np.zeros(3)):
        if mol is None:
            mol = self.mol
        return get_ovlp(mol, kpt)


class Gradients(GradientsBase):
    '''Non-relativistic Gamma-point restricted Hartree-Fock gradients'''
    def get_veff(self, mol=None, dm=None, kpt=np.zeros(3)):
        if mol is None: mol = self.mol
        if dm is None: dm = self.base.make_rdm1()
        return get_veff(self, mol, dm, kpt)

    make_rdm1e = mol_rhf.Gradients.make_rdm1e
    grad_elec = grad_elec



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Coupled cluster doubles
'''

import numpy
from pyscf.cc import ccsd
from pyscf.cc import ccsd_lambda
from pyscf.cc import ccsd_rdm

class CCD(ccsd.CCSD):
    def update_amps(self, t1, t2, eris):
        t1, t2 = ccsd.update_amps(self, t1, t2, eris)
        return numpy.zeros_like(t1), t2

    def kernel(self, t2=None, eris=None):
        nocc = self.nocc
        nvir = self.nmo - nocc
        t1 = numpy.zeros((nocc, nvir))
        ccsd.CCSD.kernel(self, t1, t2, eris)
        return self.e_corr, self.t2

    def solve_lambda(self, t2=None, l2=None, eris=None):
        if t2 is None: t2 = self.t2
        if eris is None: eris = self.ao2mo(self.mo_coeff)

        nocc = self.nocc
        nvir = self.nmo - nocc
        l1 = t1 = numpy.zeros((nocc, nvir))

        def update_lambda(mycc, t1, t2, l1, l2, eris=None, imds=None):
            l1, l2 = ccsd_lambda.update_lambda(mycc, t1, t2, l1, l2, eris, imds)
            return numpy.zeros_like(l1), l2

        self.converged_lambda, self.l1, self.l2 = \
                ccsd_lambda.kernel(self, eris, t1, t2, l1, l2,
                                   max_cycle=self.max_cycle,
                                   tol=self.conv_tol_normt,
                                   verbose=self.verbose, fupdate=update_lambda)
        return self.l2

    def make_rdm1(self, t2=None, l2=None, ao_repr=False):
        '''Un-relaxed 1-particle density matrix in MO space'''
        if t2 is None: t2 = self.t2
        if l2 is None: l2 = self.l2
        if l2 is None: l2 = self.solve_lambda(t2)

        nocc = self.nocc
        nvir = self.nmo - nocc
        l1 = t1 = numpy.zeros((nocc, nvir))

        return ccsd_rdm.make_rdm1(self, t1, t2, l1, l2, ao_repr=ao_repr)

    def make_rdm2(self, t1=None, t2=None, l1=None, l2=None, ao_repr=False):
        '''2-particle density matrix in MO space.  The density matrix is
        stored as

        dm2[p,r,q,s] = <p^+ q^+ s r>
        '''
        if t2 is None: t2 = self.t2
        if l2 is None: l2 = self.l2
        if l2 is None: l2 = self.solve_lambda(t2)

        nocc = self.nocc
        nvir = self.nmo - nocc
        l1 = t1 = numpy.zeros((nocc, nvir))

        return ccsd_rdm.make_rdm2(self, t1, t2, l1, l2, ao_repr=ao_repr)



"""
This module implements the G0W0 approximation on top of `pyscf.tdscf.rhf_slow` TDHF implementation. Unlike `gw.py`, all
integrals are stored in memory. Several variants of GW are available:

 * `pyscf.gw_slow`: the molecular implementation;
 * `pyscf.pbc.gw.gw_slow`: single-kpoint PBC (periodic boundary condition) implementation;
 * (this module) `pyscf.pbc.gw.kgw_slow_supercell`: a supercell approach to PBC implementation with multiple k-points.
   Runs the molecular code for a model with several k-points for the cost of discarding momentum conservation and using
   dense instead of sparse matrixes;
 * `pyscf.pbc.gw.kgw_slow`: a PBC implementation with multiple k-points;
"""

from pyscf.gw import gw_slow
from pyscf.lib import einsum

import numpy


# Convention for these modules:
# * IMDS contains routines for intermediates
# * kernel finds GW roots
# * GW provides a container


def corrected_moe(eri, k, p):
    """
    Calculates the corrected orbital energy.
    Args:
        eri (PhysERI): a container with electron repulsion integrals;
        k (int): the k-point index;
        p (int): orbital;

    Returns:
        The corrected orbital energy.
    """
    moe = eri.mo_energy[k][p]
    moc = eri.mo_coeff[k][:, p]
    nk = len(eri.mo_energy)
    vk = 0
    for k2 in range(nk):
        vk -= eri.ao2mo_k((
            moc[:, numpy.newaxis],
            eri.mo_coeff_full[k2][:, :eri.nocc_full[k2]],
            eri.mo_coeff_full[k2][:, :eri.nocc_full[k2]],
            moc[:, numpy.newaxis],
        ), (k, k2, k2, k)).squeeze().trace().real
    vk /= nk
    mf = eri.model
    v_mf = mf.get_veff()[k] - mf.get_j()[k]
    v_mf = einsum("i,ij,j", moc.conj(), v_mf, moc).real
    return moe + vk - v_mf


class IMDS(gw_slow.IMDS):
    orb_dims = 2

    def __init__(self, td, eri=None):
        """
        GW intermediates (k-version/supercell).
        Args:
            td: a container with TD solution;
            eri: a container with electron repulsion integrals;
        """
        gw_slow.AbstractIMDS.__init__(self, td, eri=eri)

        self.nk = len(self.td._scf.mo_energy)

        # MF
        self.nocc = sum(self.eri.nocc)
        self.o = numpy.concatenate(tuple(e[:nocc] for e, nocc in zip(self.eri.mo_energy, self.eri.nocc)))
        self.v = numpy.concatenate(tuple(e[nocc:] for e, nocc in zip(self.eri.mo_energy, self.eri.nocc)))

        # TD
        nroots, _, k1, k2, o, v = self.td.xy.shape
        self.td_xy = self.td.xy.transpose(0, 1, 2, 4, 3, 5).reshape(nroots, 2, k1*o, k2*v)
        self.td_e = self.td.e

        self.tdm = self.construct_tdm()

    def eri_ov(self, item):
        result = []
        k = numpy.arange(self.nk)
        for k1 in k:
            result.append([])
            for k2 in k:
                result[-1].append([])
                for k3 in k:
                    result[-1][-1].append([])
                    for k4 in k:
                        result[-1][-1][-1].append(self.eri.eri_ov(item, (k1, k2, k3, k4)))

        r = numpy.block(result)
        return r / len(k)

    __getitem__ = eri_ov

    def __plain_index__(self, p, spec=True):
        k, kp = p
        if kp < self.eri.nocc[k]:
            x = sum(self.eri.nocc[:k]) + kp
            if spec:
                return "o", x
            else:
                return x
        else:
            kp -= self.eri.nocc[k]
            x = sum(self.eri.nmo[:k]) - sum(self.eri.nocc[:k]) + kp
            if spec:
                return "v", x
            else:
                return x + self.nocc

    def get_rhs(self, p, components=False):
        k, kp = p
        # return self.eri.mo_energy[k][kp]
        return corrected_moe(self.eri, k, kp)

    def get_sigma_element(self, omega, p, eta, vir_sgn=1):
        return super().get_sigma_element(omega, self.__plain_index__(p, spec=False), eta, vir_sgn=vir_sgn)

    def initial_guess(self, p):
        k, kp = p
        return self.eri.mo_energy[k][kp]

    @property
    def entire_space(self):
        assert all(i == self.eri.nmo[0] for i in self.eri.nmo)
        return [numpy.arange(self.nk), numpy.arange(self.eri.nmo[0])]


kernel = gw_slow.kernel


class GW(gw_slow.GW):
    base_imds = IMDS



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xiaojie Wu <wxj6000@gmail.com>
#

'''
dispersion correction for HF and DFT
'''

import warnings
from functools import lru_cache
from pyscf.lib import logger
from pyscf import scf

# supported dispersion corrections
DISP_VERSIONS = ['d3bj', 'd3zero', 'd3bjm', 'd3zerom', 'd3op', 'd4']
XC_MAP = {'wb97m-d3bj': 'wb97m',
          'b97m-d3bj': 'b97m',
          'wb97x-d3bj': 'wb97x',
          'wb97m-v': 'wb97m',
          'b97m-v': 'b97m',
          'wb97x-v': 'wb97x'
          }

# special cases:
# - wb97x-d is not supported yet
# - wb97*-d3bj is wb97*-v with d3bj
# - wb97x-d3 is not supported yet
# - 3c method is not supported yet

# These xc functionals need special treatments
_white_list = {
    'wb97m-d3bj': ('wb97m-v', False, 'd3bj'),
    'b97m-d3bj': ('b97m-v', False, 'd3bj'),
    'wb97x-d3bj': ('wb97x-v', False, 'd3bj'),
}

# These xc functionals are not supported yet
_black_list = {
    'wb97x-d', 'wb97x-d3', 'wb97x_d', 'wb97x_d3',
    'wb97m-d3bj2b', 'wb97m-d3bjatm',
    'b97m-d3bj2b', 'b97m-d3bjatm',
}

@lru_cache(128)
def parse_dft(xc_code):
    '''
    Extract (xc, nlc, disp) from xc_code
    '''
    if not isinstance(xc_code, str):
        return xc_code, '', None
    method_lower = xc_code.lower()

    if method_lower in _black_list:
        raise NotImplementedError(f'{method_lower} is not supported yet.')

    if method_lower in _white_list:
        return _white_list[method_lower]

    if method_lower.endswith('-3c'):
        raise NotImplementedError('*-3c methods are not supported yet.')

    if '-d3' in method_lower or '-d4' in method_lower:
        xc, disp = method_lower.split('-')
    else:
        xc, disp = method_lower, None

    return xc, '', disp

@lru_cache(128)
def parse_disp(dft_method):
    '''Decode the disp parameters based on the xc code.
    Returns xc_code_for_dftd3, disp_version, with_3body

    Example: b3lyp-d3bj2b -> (b3lyp, d3bj, False)
             wb97x-d3bj   -> (wb97x, d3bj, False)
    '''
    if dft_method == 'hf':
        return 'hf', None, False

    dft_lower = dft_method.lower()
    xc, nlc, disp = parse_dft(dft_lower)
    if dft_lower in XC_MAP:
        xc = XC_MAP[dft_lower]

    if disp is None:
        return xc, None, False
    disp_lower = disp.lower()
    if disp_lower.endswith('2b'):
        return xc, disp_lower.replace('2b', ''), False
    elif disp_lower.endswith('atm'):
        return xc, disp_lower.replace('atm', ''), True
    else:
        return xc, disp_lower, False

def check_disp(mf, disp=None):
    '''Check whether to apply dispersion correction based on the xc attribute.
    If dispersion is allowed, return the DFTD3 disp version, such as d3bj,
    d3zero, d4.
    '''
    if disp is None:
        disp = mf.disp
    if disp == 0: # disp = False
        return False

    # To prevent mf.do_disp() triggering the SCF.__getattr__ method, do not use
    # method = getattr(mf, 'xc', 'hf').
    if isinstance(mf, scf.hf.KohnShamDFT):
        method = mf.xc
    else:
        # Set the disp method for both HF and MCSCF to 'hf'
        method = 'hf'
    disp_version = parse_disp(method)[1]

    if disp is None: # Using the disp version decoded from the mf.xc attribute
        if disp_version is None:
            return False
    elif disp_version is None: # Using the version specified by mf.disp
        disp_version = disp
    elif disp != disp_version:
        raise RuntimeError(f'mf.disp {disp} conflicts with mf.xc {method}')

    if disp_version not in DISP_VERSIONS:
        raise NotImplementedError
    return disp_version

def get_dispersion(mf, disp=None, with_3body=None, verbose=None):
    disp_version = check_disp(mf, disp)
    if not disp_version:
        return 0.

    try:
        from pyscf.dispersion import dftd3, dftd4
    except ImportError:
        print('dftd3 and dftd4 not available. Install them with `pip install pyscf-dispersion`')
        raise

    mol = mf.mol
    method = getattr(mf, 'xc', 'hf')
    method, _, disp_with_3body = parse_disp(method)

    if with_3body is not None:
        with_3body = disp_with_3body

    # for dftd3
    if disp_version[:2].upper() == 'D3':
        logger.info(mf, "Calc dispersion correction with DFTD3.")
        logger.info(mf, f"Parameters: xc={method}, version={disp_version}, atm={with_3body}")
        d3_model = dftd3.DFTD3Dispersion(mol, xc=method, version=disp_version, atm=with_3body)
        res = d3_model.get_dispersion()
        e_d3 = res.get('energy')
        mf.scf_summary['dispersion'] = e_d3
        return e_d3

    # for dftd4
    elif disp_version[:2].upper() == 'D4':
        logger.info(mf, "Calc dispersion correction with DFTD4.")
        logger.info(mf, f"Parameters: xc={method}, atm={with_3body}")
        d4_model = dftd4.DFTD4Dispersion(mol, xc=method, atm=with_3body)
        res = d4_model.get_dispersion()
        e_d4 = res.get('energy')
        mf.scf_summary['dispersion'] = e_d4
        return e_d4
    else:
        raise RuntimeError(f'dispersion correction: {disp_version} is not supported.')
